"""Configuration definitions."""

from __future__ import annotations

import copy
import re

from dataclasses import InitVar
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import NewType
from typing import TypeVar

from ansible_navigator.utils.functions import expand_path
from ansible_navigator.utils.functions import oxfordcomma


if TYPE_CHECKING:
    from collections.abc import Iterable

    from .navigator_configuration import Internals
    from .navigator_post_processor import NavigatorPostProcessor


def version_added_sanity_check(version: str) -> None:
    """Check if a version string is valid.

    Args:
        version: The version string to check
    """
    re_version = re.compile(r"^v\d+\.\d+$")
    assert (  # noqa:S101
        re_version.match(version) is not None
    ), "Version must be in the form of v{major}.{minor}"


class Constants(Enum):
    """Mapping some constants to friendly text."""

    ALL = "All"
    ANSIBLE_CFG = "Ansible configuration file"
    AUTO = "Automatically determined"
    DEFAULT_CFG = "Defaults"
    ENVIRONMENT_VARIABLE = "Environment variable"
    NONE = "None"
    NOT_SET = "Not set"
    PREVIOUS_CLI = "Previous cli command"
    SAME_SUBCOMMAND = (
        "Used to determine if an entry should be used when"
        " applying previous cli common entries, this indicates"
        " that it will only be used if the subcommand is the same"
    )
    SEARCH_PATH = "Search path"
    SENTINEL = "Indicates a non value"
    TEST = "Indicates a value set from a test"
    USER_CFG = "Settings file"
    USER_CLI = "Command line"

    def __str__(self) -> str:
        """Use the value when presented as a string.

        Returns:
            The value as type str
        """
        return str(self.value)


# The following are ordered to build up to an ApplicationConfiguration


@dataclass
class CliParameters:
    """An object to hold the CLI parameters."""

    action: str | None = None
    const: bool | str | None = None
    long_override: str | None = None
    nargs: str | None = None
    positional: bool = False
    short: str | None = None
    metavar: str | None = None

    def long(self, name_dashed: str) -> str:
        """Provide a cli long parameter.

        Args:
            name_dashed: The dashed name of the parent settings entry

        Returns:
            The long cli parameter
        """
        long = self.long_override or f"--{name_dashed}"
        return long


@dataclass
class SettingsEntryValue:
    """An object to store a value."""

    #: The default value for the entry
    default: Any = Constants.NOT_SET
    #: The current, effective value for the entry
    current: Any = Constants.NOT_SET
    #: Provide a specific default value to be used in the schema
    schema_default: str | Constants = Constants.NOT_SET
    #: Indicates where the current value came from
    source: Constants = Constants.NOT_SET

    @property
    def is_default(self) -> bool:
        """Determine if the current value is the default value.

        Returns:
            Indication of if the current is the default
        """
        result = self.default == self.current
        return result

    @property
    def resolved(self) -> SettingsEntryValue:
        """Transform this entry to an entry without internal constants.

        This would typically be used when the attributes need to be presented to the user.
        Constants are resolved to their value.

        Returns:
            An entry without internal constants for attributes
        """
        new_entry = copy.deepcopy(self)

        if isinstance(self.current, Constants):
            new_entry.current = self.current.value

        if isinstance(self.default, Constants):
            new_entry.default = self.default.value

        return new_entry


@dataclass
class SettingsEntry:
    # pylint: disable=too-many-instance-attributes
    """One entry in the configuration."""

    #: The reference name for the entry
    name: str
    #: A short description used for the argparse help
    short_description: str
    #: The value for the entry
    value: SettingsEntryValue
    #: Version added
    version_added: str
    #: Indicates if this should be applied to future CLIs parsed
    apply_to_subsequent_cli: Constants = Constants.ALL
    #: Indicates if this can be changed after initialization
    change_after_initial: bool = True
    #: The possible values for this entry
    choices: Iterable[bool | str] = field(default_factory=list)
    #: Argparse specific params
    cli_parameters: CliParameters | None = None
    #: Post process in normal (alphabetical) order or wait until after first pass
    delay_post_process: bool = False
    #: Override the default, generated environment variable
    environment_variable_override: str | None = None
    #: Over the default settings file path, dot delimited representation in tree
    environment_variable_split_char: str = ","
    #: The character used to split an environment variable value into a list
    settings_file_path_override: str | None = None
    #: Subcommand this should this be used for
    subcommands: list[str] | Constants = Constants.ALL
    #: Does this hold the name of the active subcommand
    subcommand_value: bool = False

    def __post_init__(self) -> None:
        """Perform post initialization actions."""
        version_added_sanity_check(self.version_added)

    def environment_variable(self, prefix: str = "") -> str:
        """Generate an effective environment variable for this entry.

        Args:
            prefix: The prefix for environmental variable

        Returns:
            Environmental variable with prefix prepended
        """
        if self.environment_variable_override is not None:
            env_var = self.environment_variable_override
        else:
            env_var = f"{prefix}_{self.name.replace('--', '')}"
        env_var = env_var.replace("-", "_").upper()
        return env_var

    @property
    def invalid_choice(self) -> str:
        """Generate an invalid choice message for this entry.

        Raises:
            ValueError: If source is not set for that entry

        Returns:
            Constructed message
        """
        name = self.name.replace("_", "-")
        if self.value.source is Constants.NOT_SET:
            msg = f"Current source not set for {self.name}"
            raise ValueError(msg)

        choices = [str(choice).lower() for choice in self.choices]
        current = self.value.current
        if isinstance(current, list):
            choices_str = oxfordcomma(choices, "and/or")
            current = oxfordcomma(current, "and")
            prefix = f"The setting '{name}' must be one or more of"
        else:
            choices_str = oxfordcomma(choices, "or")
            prefix = f"The setting '{name}' must be one of"

        source = self.value.source.value
        msg = f"{prefix} {choices_str}, but set as '{current}'. ({source})"
        return msg

    @property
    def name_dashed(self) -> str:
        """Generate a dashed version of the name.

        Returns:
            Dashed version of the name
        """
        return self.name.replace("_", "-")

    def settings_file_path(self, prefix: str) -> str:
        """Generate an effective settings file path for this entry.

        Args:
            prefix: The prefix for the settings file path

        Returns:
            Settings file path
        """
        prefix_str = f"{prefix}." if prefix else prefix

        if self.settings_file_path_override is not None:
            sfp = f"{prefix_str}{self.settings_file_path_override}"
        else:
            sfp = f"{prefix_str}{self.name}"

        return sfp.replace("_", "-")


@dataclass(frozen=True)
class SubCommand:
    """An object to hold a subcommand."""

    name: str
    description: str
    version_added: str
    epilog: str | None = None

    def __post_init__(self) -> None:
        """Perform post initialization actions."""
        version_added_sanity_check(self.version_added)


@dataclass
class ApplicationConfiguration:
    # pylint: disable=too-many-instance-attributes
    """The main object for storing an application config."""

    application_version: Constants | str
    entries: list[SettingsEntry]
    internals: Internals
    post_processor: NavigatorPostProcessor | None
    subcommands: list[SubCommand]

    application_name: str = ""
    original_command: list[str] = field(default_factory=list)

    initial: Any = None

    @property
    def application_name_dashed(self) -> str:
        """Generate a dashed version of the application name.

        Returns:
            Application name dashed
        """
        return self.application_name.replace("_", "-")

    def _get_by_name(self, name: str, kind: str) -> SettingsEntry:
        """Retrieve a settings entry by name.

        Args:
            name: The name of the entry
            kind: The kind of entry to retrieve

        Returns:
            The settings entry

        Raises:
            KeyError: If the entry is not found
        """
        try:
            return next(entry for entry in super().__getattribute__(kind) if entry.name == name)
        except StopIteration as exc:
            raise KeyError(name) from exc

    def __getattribute__(self, attr: str) -> Any:
        """Return a matching entry or the default from super.

        Args:
            attr: The attribute to get

        Returns:
            Either the matching entry or default from super
        """
        try:
            return super().__getattribute__("_get_by_name")(attr, "entries").value.current
        except (AttributeError, KeyError):
            return super().__getattribute__(attr)

    def entry(self, name: str) -> SettingsEntry:
        """Retrieve a configuration entry by name.

        Args:
            name: The name of the entry

        Returns:
            Configuration entry name
        """
        return self._get_by_name(name, "entries")

    def subcommand(self, name: str) -> SubCommand:
        """Retrieve a configuration subcommand by name.

        Args:
            name: The name of the subcommand

        Returns:
            Configuration subcommand name

        Raises:
            TypeError: if there is a typing error
        """
        command = self._get_by_name(name, "subcommands")
        if not isinstance(command, SubCommand):
            raise TypeError
        return command


# The following are ordered to build up to an VolumeMount


class VolumeMountOption(Enum):
    """Options that can be tagged on to the end of volume mounts.

    Usually these are used for things like selinux relabeling, but there are
    some other valid options as well, which can and should be added here as
    needed. See ``man podman-run`` and ``man docker-run`` for valid choices and
    keep in mind that we support both runtimes.
    """

    # Overlay
    OVERLAY = "O"

    # Read Only
    ro = "ro"

    # Read Write
    rw = "rw"

    # Relabel as private
    Z = "Z"

    # Relabel as shared.
    z = "z"


V = TypeVar("V", bound="VolumeMount")


class VolumeMountError(Exception):
    """Custom exception raised when building VolumeMounts."""


@dataclass(frozen=True)
class VolumeMount:
    """Describes EE volume mounts.

    Attributes:
        fs_source: The source file system path of the volume mount.
        fs_destination: The destination file system path in the container for the volume mount.
        settings_entry: The name of the settings entry requiring this volume mount.
        source: The settings source for this volume mount.
        options_string: Comma delimited options.
        options: Options for the bind mount.
    """

    fs_source: str
    fs_destination: str
    settings_entry: str
    source: Constants = field(compare=False)
    options_string: InitVar[str]
    options: tuple[VolumeMountOption, ...] = ()

    def __post_init__(self, options_string: str) -> None:
        """Post process the ``VolumeMount`` and perform sanity checks.

        Raises:
            VolumeMountError: When a viable VolumeMount cannot be
                created

        Args:
            options_string: Option entries in as type string
        """
        errors = []

        # Ensure each is a string
        if not isinstance(self.fs_source, str):
            errors.append(f"Source: '{self.fs_source}' is not a string.")
        if not isinstance(self.fs_destination, str):
            errors.append(f"Destination: '{self.fs_destination}' is not a string.")
        if not isinstance(options_string, str):
            errors.append(f"Options: '{options_string}' is not a string.")

        # Ensure source and dest are not empty
        if self.fs_source == "":
            errors.append("Source not provided.")
        if self.fs_destination == "":
            errors.append("Destination not provided.")

        # Exit early for errors
        if errors:
            raise VolumeMountError(" ".join(errors))

        # Resolve the source and destination
        # frozen, cannot use simple assignment to initialize fields, and must use:
        object.__setattr__(self, "fs_source", expand_path(self.fs_source))
        object.__setattr__(self, "fs_destination", expand_path(self.fs_destination))

        # Source must exist
        if not Path(self.fs_source).exists():
            error_msg = f"Source: '{self.fs_source}' does not exist."
            raise VolumeMountError(error_msg)

        # Validate and populate _options
        if options_string == "":
            return

        options = []
        option_values = [o.value for o in VolumeMountOption]
        for option in options_string.split(","):
            if option not in option_values:
                errors.append(
                    f"Unrecognized option: '{option}',"
                    f" available options include"
                    f" {oxfordcomma(option_values, condition='and/or')}.",
                )
            else:
                options.append(VolumeMountOption(option))

        if errors:
            raise VolumeMountError(" ".join(errors))

        unique = sorted(set(options), key=options.index)
        object.__setattr__(self, "options", tuple(unique))

    def to_string(self) -> str:
        """Render the volume mount in a way that (docker|podman) understands.

        Returns:
            File system source and system path for volume mount
        """
        out = f"{self.fs_source}:{self.fs_destination}"
        if self.options:
            joined_opts = ",".join(o.value for o in self.options)
            out += f":{joined_opts}"
        return out


# The following are ordered to build up to a Mode


class Mode(Enum):
    """An enum to restrict mode type.

    Attributes:
        STDOUT: Represents the "stdout" mode.
        INTERACTIVE: Represents the "interactive" mode.
    """

    STDOUT = "stdout"
    INTERACTIVE = "interactive"


@dataclass
class ModeChangeRequest:
    """Data structure to contain a mode change request by a settings entry.

    Attributes:
        entry: The entry making the request.
        mode: The desired mode.
    """

    entry: str
    mode: Mode


@dataclass
class PaeChangeRequest:
    """Data structure to hold playbook artifact change request by a settings entry.

    Attributes:
        entry: The entry making the request.
        playbook_artifact_enable: The desired value for playbook_artifact_enable.
    """

    entry: str
    playbook_artifact_enable: bool


# and some common ones

# A type used for the settings as a dictionary
SettingsFileType = NewType(
    "SettingsFileType",
    dict[str, bool | dict[Any, Any] | int | str | list[Any]],
)

# A type used to describe a schema file for the settings
SettingsSchemaType = NewType(
    "SettingsSchemaType",
    dict[str, dict[str, bool | dict[Any, Any] | int | str | list[Any]]],
)



"""Common definitions for a version migration."""

from __future__ import annotations

import contextlib

from enum import Enum
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import TypeVar

from ansible_navigator.utils.ansi import COLOR
from ansible_navigator.utils.ansi import changed
from ansible_navigator.utils.ansi import failed
from ansible_navigator.utils.ansi import subtle
from ansible_navigator.utils.ansi import working


if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


class MigrationType(Enum):
    """Enum for the type of migration.

    Attributes:
        SETTINGS_FILE: Migration of the settings file.
    """

    SETTINGS_FILE = "settings"


T = TypeVar("T")


class MigrationStep(Generic[T]):
    """Data class for a migration step."""

    def __init__(self, name: str) -> None:
        """Initialize a migration step.

        Args:
            name: The name of the migration step
        """
        self.name: str = name
        """The name of the migration step"""
        self.needed: bool = False
        """Whether the migration step is needed"""
        self.function_name: str | None = None
        """The name of the function to call"""

    def print_start(self) -> None:
        """Output start information to the console."""
        message = f"Migrating '{self.name}'"
        information = f"{message:.<60}"
        working(color=COLOR, message=information)

    def print_failed(self) -> None:
        """Output fail information to the console."""
        message = f"Migration of '{self.name}'"
        information = f"{message:.<60}Failed"
        failed(color=COLOR, message=information)

    def print_updated(self) -> None:
        """Output updated information to the console."""
        message = f"Migration of '{self.name}'"
        information = f"{message:.<60}Updated"

        changed(color=COLOR, message=information)

    def print_not_needed(self) -> None:
        """Output not needed information to the console."""
        message = f"Migration of '{self.name}'"
        information = f"{message:.<60}Not needed"

        subtle(color=COLOR, message=information)

    @classmethod
    def register(cls: T, migration_step: MigrationStep[Any]) -> Callable[..., Any]:
        """Register the migration step.

        Args:
            migration_step: The migration step to register

        Returns:
            The registered migration step
        """

        def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
            """Add the dunder collector to the func.

            Args:
                func: The function to decorate

            Returns:
                The decorated function
            """
            migration_step.function_name = func.__name__
            func.__migration_step__ = migration_step  # type: ignore[attr-defined]
            return func

        return wrapper


class Migration:
    """Data class for a migration.

    Attributes:
        name: The name of the migration.
    """

    name = "Migration base class"

    def __init__(self) -> None:
        """Initialize the migration."""
        self.check: bool = False
        """Whether the migration is needed."""
        self.settings_file_path: Path
        """The path of the settings file."""
        self.was_needed: bool = False
        """Whether the migration was needed."""

    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        """Register the migration steps.

        Args:
            *args: Positional arguments
            **kwargs: Keyword arguments
        """
        super().__init_subclass__(*args, **kwargs)
        migrations.append(cls)

    @property
    def migration_steps(self) -> tuple[MigrationStep[Any], ...]:
        """Return the registered diagnostics.

        Returns:
            The registered diagnostics
        """
        steps: list[MigrationStep[Any]] = []
        for func_name in vars(self.__class__):
            if func_name.startswith("_"):
                continue
            if hasattr(getattr(self, func_name), "__migration_step__"):
                step = getattr(self, func_name).__migration_step__
                steps.append(step)
        return tuple(steps)

    @property
    def needed_now(self) -> bool:
        """Return whether the migration is needed.

        Returns:
            Whether the migration is needed
        """
        return any(step.needed for step in self.migration_steps)

    def run(self, *args: Any, **kwargs: Any) -> None:
        """Run the migration.

        Args:
            *args: The positional arguments
            **kwargs: The keyword arguments
        """

    def run_step(self, step: MigrationStep[Any], *args: Any, **kwargs: Any) -> None:
        """Run the migration step.

        Args:
            step: The migration step to run
            *args: The positional arguments
            **kwargs: The keyword arguments
        """
        if not isinstance(step.function_name, str):
            return
        function = getattr(self, step.function_name)
        if self.check:
            try:
                step.needed = function(*args, **kwargs)
            except Exception:  # noqa: BLE001
                step.needed = False
            return

        if not step.needed:
            step.print_not_needed()
            return

        step.print_start()
        with contextlib.suppress(Exception):
            step.needed = function(*args, **kwargs)

        if step.needed:
            step.print_failed()
        else:
            step.print_updated()

    def run_steps(self, *args: Any, **kwargs: Any) -> None:
        """Run all registered migration steps.

        Args:
            *args: The positional arguments
            **kwargs: The keyword arguments
        """
        for step in self.migration_steps:
            self.run_step(step, *args, **kwargs)


Migrations = list[type[Migration]]
migrations: Migrations = []



"""``:save`` command implementation."""

import logging

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action:
    """``:save`` command implementation."""

    KEGEX = r"^s(?:ave)?\s(?P<filename>.*)$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:save`` action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the ``:save`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance
        """
        self._logger.debug("save requested")
        filename = interaction.action.match.groupdict()["filename"]
        app.write_artifact(filename)



"""Functionality to perform json schema validation."""

from __future__ import annotations

import json

from collections import deque
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any

from jsonschema import SchemaError
from jsonschema import ValidationError
from jsonschema.validators import validator_for

from .definitions import ExitMessage


if TYPE_CHECKING:
    from collections.abc import Sequence


def to_path(schema_path: Sequence[Any] | str) -> str:
    """Flatten a path to a dot delimited string.

    Args:
        schema_path: The schema path

    Returns:
        The dot delimited path
    """
    queue = schema_path if isinstance(schema_path, deque) else deque(schema_path)
    return ".".join(str(index) for index in queue)


def json_path(absolute_path: Sequence[Any]) -> str:
    """Flatten a data path to a dot delimited string.

    Args:
        absolute_path: The path

    Returns:
        The dot delimited string
    """
    path = "$"
    for elem in absolute_path:
        if isinstance(elem, int):
            path += "[" + str(elem) + "]"
        else:
            path += "." + elem
    return path


@dataclass
class JsonSchemaError:
    # pylint: disable=too-many-instance-attributes
    """Data structure to hold a json schema validation error."""

    message: str
    data_path: str
    json_path: str
    schema_path: str
    relative_schema: str
    expected: bool | int | str
    validator: str
    found: str

    def to_friendly(self) -> str:
        """Provide a friendly explanation of the error.

        Returns:
            The error message
        """
        return f"In '{self.data_path}': {self.message}."

    def to_exit_message(self) -> ExitMessage:
        """Provide an exit message for a schema validation failure.

        Returns:
            The exit message
        """
        return ExitMessage(message=self.to_friendly())


def validate(schema: str | dict[str, Any], data: dict[str, Any]) -> list[JsonSchemaError]:
    """Validate some data against a JSON schema.

    Args:
        schema: the JSON schema to use for validation
        data: The data to validate

    Returns:
        Any errors encountered

    Raises:
        TypeError: If there is a typing error.
    """
    errors: list[JsonSchemaError] = []

    if isinstance(schema, str):
        schema = json.loads(schema)
    if isinstance(schema, (bool, str)):
        msg = "Unexpected schema data."
        raise TypeError(msg)
    validator = validator_for(schema)
    try:
        validator.check_schema(schema)
    except SchemaError as exc:
        error = JsonSchemaError(
            message=str(exc),
            data_path="schema sanity check",
            json_path="",
            schema_path="",
            relative_schema="",
            expected="",
            validator="",
            found="",
        )
        errors.append(error)
        return errors

    validation_errors = sorted(validator(schema).iter_errors(data), key=lambda e: e.path)

    if not validation_errors:
        return errors

    for validation_error in validation_errors:
        if isinstance(validation_error, ValidationError):
            error = JsonSchemaError(
                message=validation_error.message,
                data_path=to_path(validation_error.absolute_path),
                json_path=json_path(validation_error.absolute_path),
                schema_path=to_path(validation_error.relative_schema_path),
                relative_schema=str(validation_error.schema),
                expected=str(validation_error.validator_value),
                validator=str(validation_error.validator),
                found=str(validation_error.instance),
            )
            errors.append(error)
    return errors



"""Logging initialization."""

from __future__ import annotations

import datetime
import logging
import zoneinfo

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from .configuration_subsystem import Constants


if TYPE_CHECKING:
    from .configuration_subsystem.definitions import ApplicationConfiguration


logger = logging.getLogger("ansible_navigator")


class Formatter(logging.Formatter):
    """Format a logging timestamp using a time zone."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the logging formatter.

        Args:
            *args: The arguments
            **kwargs: The keyword arguments
        """
        self._time_zone = kwargs.pop("time_zone")
        super().__init__(*args, **kwargs)

    def formatTime(  # noqa: N802
        self,
        record: logging.LogRecord,
        _datefmt: str | None = None,
    ) -> str:
        """Format the log timestamp.

        Args:
            record: The log record
            _datefmt: The optional date format

        Returns:
            The timestamp
        """
        if self._time_zone == "local":
            return (
                datetime.datetime.fromtimestamp(record.created, datetime.timezone.utc)
                .astimezone()
                .isoformat()
            )
        if self._time_zone == "UTC":
            return datetime.datetime.fromtimestamp(
                record.created,
                tz=datetime.timezone.utc,
            ).isoformat()
        try:
            return datetime.datetime.fromtimestamp(
                record.created,
                tz=zoneinfo.ZoneInfo(self._time_zone),
            ).isoformat()
        except zoneinfo.ZoneInfoNotFoundError:
            return datetime.datetime.fromtimestamp(
                record.created,
                tz=datetime.timezone.utc,
            ).isoformat()


def setup_logger(args: ApplicationConfiguration) -> None:
    """Set up the logger.

    Args:
        args: The CLI args
    """
    if Path(args.log_file).exists() and args.log_append is False:
        Path(args.log_file).unlink()
    handler = logging.FileHandler(args.log_file)

    time_zone = args.entry("time_zone").value.current
    # When the configuration is rolled back, the time_zone will be C.NOT_SET
    if isinstance(args.time_zone, Constants):
        time_zone = args.entry("time_zone").value.default

    formatter = Formatter(
        fmt="%(asctime)s %(levelname)s '%(name)s.%(funcName)s' %(message)s",
        time_zone=time_zone,
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    log_level = getattr(logging, args.log_level.upper())
    logger.setLevel(log_level)
    logger.info("New %s instance, logging initialized", args.application_name_dashed)

    # set ansible-runner logs
    runner_logger = logging.getLogger("ansible-runner")
    runner_logger.setLevel(log_level)
    runner_logger.addHandler(handler)
    logger.info("New ansible-runner instance, logging initialized")



"""A text input field."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any

from .form_handler_button import FormHandlerButton
from .validators import FieldValidators


if TYPE_CHECKING:
    from collections.abc import Callable

    from .curses_window import Window
    from .form_defs import FieldValidationStates


@dataclass
class FieldButton:
    """A text input field."""

    name: str
    text: str
    disabled: bool = True
    pressed: bool = False
    color: int = 0
    window_handler = FormHandlerButton
    validator: Callable[..., Any] = FieldValidators.none
    win: Window | None = None

    @property
    def full_prompt(self) -> str:
        """No default to add into the prompt for checkbox.

        Returns:
            Empty string
        """
        return ""

    def validate(self, response: FieldValidationStates) -> None:
        """Validate this instance.

        Args:
            response: List of field states for validation.
        """
        validation = self.validator(response)
        if validation.error_msg:
            self.disabled = True
        else:
            self.disabled = False

    def conditional_validation(self, response: FieldValidationStates) -> None:
        """Conditional validation used for form validation.

        Args:
            response: List of field states for validation.
        """
        self.validate(response)



"""Get one line of text input."""

from __future__ import annotations

import curses

from curses import ascii as curses_ascii
from typing import TYPE_CHECKING

from .curses_defs import CursesLine
from .curses_defs import CursesLinePart
from .curses_window import CursesWindow


if TYPE_CHECKING:
    from ansible_navigator.action_runner import Window
    from ansible_navigator.ui_framework.ui_config import UIConfig

    from .field_button import FieldButton


class FormHandlerButton(CursesWindow):
    """Handle form button."""

    def __init__(self, screen: Window, ui_config: UIConfig) -> None:
        """Initialize the handler for a form button.

        Args:
            screen: A curses window
            ui_config: The current user interface configuration
        """
        super().__init__(ui_config=ui_config)
        self._form_field: FieldButton | None = None
        self._form_fields: list[FieldButton] | None = None
        self._screen = screen

    def populate(self) -> None:
        """Populate the window with the button.

        Raises:
            RuntimeError: if there is a runtime error
        """
        if not self._form_field:
            msg = "_form_field not initialized"
            raise RuntimeError(msg)
        color = 8 if self._form_field.disabled is True else self._form_field.color

        if self._ui_config.color is False:
            text = f"[{self._form_field.text.upper()}]"
        else:
            text = self._form_field.text

        clp_button = CursesLinePart(0, text, color, curses.A_STANDOUT)
        self._add_line(self.win, 0, CursesLine((clp_button,)))

    def handle(self, idx: int, form_fields: list[FieldButton]) -> tuple[FieldButton, int]:
        """Handle the check box field.

        Args:
            form_fields: List of fields
            idx: Index to retrieve specific field

        Returns:
            Field and input from said field
        """
        self._form_fields = form_fields
        self._form_field = form_fields[idx]
        self.populate()

        while True:
            char = self.win.getch()

            if char in [curses.KEY_RESIZE, curses_ascii.TAB]:
                break

            if char in [curses_ascii.NL, curses_ascii.CR]:
                if not self._form_field.disabled:
                    self._form_field.pressed = True
                break
        return self._form_field, char



"""Some utilities that are specific to ansible_navigator."""

from __future__ import annotations

import ast
import datetime
import decimal
import html
import logging
import os
import re
import shlex
import shutil
import zoneinfo

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from jinja2 import Environment
from jinja2 import StrictUndefined
from jinja2 import TemplateError

from .definitions import GOLDEN_RATIO
from .definitions import ExitMessage
from .definitions import LogMessage


if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Mapping


logger = logging.getLogger(__name__)


def oxfordcomma(listed: Iterable[bool | str | Path], condition: str) -> str:
    """Format a list into a sentence.

    Args:
        listed: List of string entries to modify
        condition: String to splice into string, usually 'and'

    Returns:
        Modified string
    """
    listed = [f"'{entry!s}'" for entry in listed]
    if len(listed) == 0:
        return ""
    if len(listed) == 1:
        return str(listed[0])
    if len(listed) == 2:
        return f"{listed[0]} {condition} {listed[1]}"
    return f"{', '.join(str(x) for x in listed[:-1])} {condition} {listed[-1]}"


def expand_path(path: str | Path) -> Path:
    """Resolve a path.

    Args:
        path: The file path to resolve

    Returns:
        Resolved file path
    """
    _path = Path(os.path.expandvars(path))
    _path = _path.expanduser()
    return _path.resolve()


def check_for_ansible() -> tuple[list[LogMessage], list[ExitMessage]]:
    """Check for the ansible-playbook command, runner will need it.

    Returns:
        Exit messages if not found, messages if found
    """
    messages: list[LogMessage] = []
    exit_messages: list[ExitMessage] = []
    ansible_location = shutil.which("ansible-playbook")
    if not ansible_location:
        msg_parts = [
            "The 'ansible-playbook' command could not be found or was not executable,",
            "ansible is required when running without an Ansible Execution Environment.",
            "Try one of",
            "     'pip install ansible-base'",
            "     'pip install ansible-core'",
            "     'pip install ansible'",
            "or simply",
            "     '-ee' or '--execution-environment'",
            "to use an Ansible Execution Environment",
        ]
        exit_messages.append(ExitMessage(message="\n".join(msg_parts)))
        return messages, exit_messages
    message = f"ansible-playbook found at {ansible_location}"
    messages.append(LogMessage(level=logging.INFO, message=message))
    return messages, exit_messages


def check_playbook_type(playbook: str) -> str:
    """Determine if given playbook is a file, fqcn playbook or something else.

    Note: These checks are added to directly access a playbook provided by a collection
    i.e. `namespace.collection.playbook_name` fqcn format.

    Args:
        playbook: given playbook

    Returns:
        playbook type
    """
    playbook_type = "file"
    playbook_path = str(playbook)
    if Path(playbook_path).exists() is False:
        playbook_type = "missing"
    if Path(playbook_path).exists() is False and len(playbook_path.split(".")) >= 3:
        playbook_type = "fqcn"
    return playbook_type


def clear_screen() -> None:
    """Print blank lines on the screen, preserving scrollback.

    Note: In certain cases, xterm.js based terminals show stdout
    under the initial curses output.

    Rather than issuing :command:`clear`, print blank lines to
    preserve the user's scrollback buffer.
    """
    affected_terminals = ["vscode"]
    if os.environ.get("TERM_PROGRAM") in affected_terminals:
        for _line in range(shutil.get_terminal_size().lines):
            print()


def console_width() -> int:
    """Get a console width based on common screen widths.

    Returns:
        The console width
    """
    width = shutil.get_terminal_size().columns
    if width <= 80:
        return width
    if width <= 132:
        return max(80, round_half_up(width / GOLDEN_RATIO))
    return 132


# Replace this with something type-safe.
def dispatch(obj: object, replacements: tuple[tuple[str, str], ...]) -> object:
    """Make the replacement based on type.

    Args:
        obj: An obj in which replacements will be made
        replacements: The things to replace

    Returns:
        Variable obj
    """
    if isinstance(obj, dict):
        obj = {k: dispatch(v, replacements) for k, v in obj.items()}
    elif isinstance(obj, list):
        obj = [dispatch(l, replacements) for l in obj]  # noqa: E741
    elif isinstance(obj, str):
        for replacement in replacements:
            obj = obj.replace(replacement[0], replacement[1])
    return obj


def escape_moustaches(obj: Mapping[Any, Any]) -> Mapping[Any, Any]:
    """Escape moustaches.

    Args:
        obj: Variable that may contain moustaches

    Returns:
        The obj with replacements made

    Raises:
        TypeError: If there is a typing error.
    """
    replacements = (("{", "U+007B"), ("}", "U+007D"))
    result = dispatch(obj, replacements)
    if not isinstance(result, dict):
        raise TypeError
    return result


def environment_variable_is_file_path(
    env_var: str,
    kind: str,
) -> tuple[list[LogMessage], list[ExitMessage], str | None]:
    """Check if a given environment variable is a viable file path, and if so, return that path.

    Args:
        env_var: Environment variable to check for a file path
        kind: Type of file

    Returns:
        Log messages and file path
    """
    messages: list[LogMessage] = []
    exit_messages: list[ExitMessage] = []
    file_path = None
    candidate_path = os.environ.get(env_var)
    if candidate_path is None:
        messages.append(LogMessage(level=logging.DEBUG, message=f"No {kind} file set by {env_var}"))
    else:
        message = f"Found a {kind} file at {candidate_path} set by {env_var}"
        messages.append(LogMessage(level=logging.DEBUG, message=message))
        if Path(candidate_path).is_file() and Path(candidate_path).exists():
            file_path = candidate_path
            message = f"{kind.capitalize()} file at {file_path} set by {env_var} is viable"
            messages.append(LogMessage(level=logging.DEBUG, message=message))
            exp_path = str(expand_path(file_path))
            if exp_path != file_path:
                message = f"{kind.capitalize()} resolves to {exp_path}"
                messages.append(LogMessage(level=logging.DEBUG, message=message))
                file_path = exp_path
        else:
            message = f"{env_var} set as {candidate_path} but not valid"
            messages.append(LogMessage(level=logging.DEBUG, message=message))
    return messages, exit_messages, file_path


def find_settings_file() -> tuple[list[LogMessage], list[ExitMessage], Path | None]:
    """Find the settings file.

    Find the file at ./ansible-navigator.(.yml,.yaml,.json),
    or ~/.ansible-navigator.(.yml,.yaml,.json).

    Returns:
        Log messages and correct settings file to use
    """
    messages: list[LogMessage] = []
    exit_messages: list[ExitMessage] = []
    allowed_extensions = ["yml", "yaml", "json"]
    potential_paths: list[Path] = []
    found_files: list[Path] = []

    settings_file_home = Path.home() / ".ansible-navigator"
    settings_file_current = Path.cwd() / "ansible-navigator"
    potential_paths.append(settings_file_home)
    potential_paths.append(settings_file_current)

    for path in potential_paths:
        message = f"Looking in {path}"
        messages.append(LogMessage(level=logging.DEBUG, message=message))

        candidates: list[Path] = []
        for ext in allowed_extensions:
            p = Path(f"{path}.{ext}")
            candidates.append(p)
        message = f"Looking for {oxfordcomma(candidates, 'and')}"
        messages.append(LogMessage(level=logging.DEBUG, message=message))

        found = [file for file in candidates if file.exists()]

        message = f"Found {len(found)}: {oxfordcomma(found, 'and')}"
        messages.append(LogMessage(level=logging.DEBUG, message=message))

        if len(found) > 1:
            exit_msg = f"Only one file among {oxfordcomma(candidates, 'and')}"
            exit_msg += f" should be present in {path}"
            exit_msg += f" Found: {oxfordcomma(found, 'and')}"
            exit_messages.append(ExitMessage(message=exit_msg))
            return messages, exit_messages, None
        found_files.extend(found)

    message = f"All found {len(found_files)}: {oxfordcomma(found_files, 'and')}"
    messages.append(LogMessage(level=logging.DEBUG, message=message))

    use = found_files[-1] if found_files else None
    message = f"Using: {use}"
    messages.append(LogMessage(level=logging.DEBUG, message=message))

    return messages, exit_messages, use


def flatten_list(data_list: list[Any]) -> list[Any]:
    """Flatten a list of lists.

    Args:
        data_list: List to flatten

    Returns:
        Flattened list
    """
    if isinstance(data_list, list):
        return [a for i in data_list for a in flatten_list(i)]
    return [data_list]


def generate_cache_path(app_name: str) -> Path:
    """Return the path to the cache directory.

    Args:
        app_name: Name of application - currently ansible_navigator

    Returns:
        Path to the cache directory
    """
    cache_home = os.environ.get("XDG_CACHE_HOME", f"{Path.home()}/.cache")
    return Path(cache_home) / app_name


def divmod_int(numerator: float, denominator: float) -> tuple[int, int]:
    """Return the result of divmod, as a tuple of integers.

    Args:
        numerator: Numerator for divmod
        denominator: Denominator for divmod

    Returns:
        Quotient and remainder of divmod
    """
    quotient, remainder = divmod(numerator, denominator)
    return int(quotient), int(remainder)


def human_time(seconds: float) -> str:
    """Convert seconds into human readable 00d00h00m00s format.

    Args:
        seconds: Time in seconds

    Returns:
        Human readable conversion of seconds
    """
    sign_string = "-" if seconds < 0 else ""
    seconds = abs(int(seconds))
    days, seconds = divmod_int(seconds, 86400)
    hours, seconds = divmod_int(seconds, 3600)
    minutes, seconds = divmod_int(seconds, 60)
    if days > 0:
        return f"{sign_string!s}{days:d}d{hours:d}h{minutes:d}m{seconds:d}s"
    if hours > 0:
        return f"{sign_string!s}{hours:d}h{minutes:d}m{seconds:d}s"
    if minutes > 0:
        return f"{sign_string!s}{minutes:d}m{seconds:d}s"
    return f"{sign_string!s}{seconds:d}s"


def is_jinja(string: str) -> bool:
    """Determine if a string is a Jinja2 template.

    Args:
        string: The string to check.

    Returns:
        True if the string is a Jinja2 template, False otherwise.
    """
    try:
        return string.index("{{") < string.index("}}")
    except ValueError:
        return False


def now_iso(time_zone: str) -> str:
    """Return the current time as an ISO 8601 formatted string, given a time zone.

    Args:
        time_zone: The IANA timezone name or local

    Returns:
        The ISO 8601 formatted time zone string
    """
    if time_zone == "local":
        return datetime.datetime.now(tz=datetime.timezone.utc).astimezone().isoformat()
    if time_zone == "UTC":
        return datetime.datetime.now(tz=datetime.timezone.utc).isoformat()
    try:
        return datetime.datetime.now(tz=zoneinfo.ZoneInfo(time_zone)).isoformat()
    except zoneinfo.ZoneInfoNotFoundError:
        logger.exception("The time zone '%s' could not be found. Using UTC.", time_zone)
        return datetime.datetime.now(tz=datetime.timezone.utc).isoformat()


PASCAL_REGEX = re.compile("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")


def pascal_to_snake(obj: object) -> object:
    """Convert a pascal cased object into a snake cased object recursively.

    Args:
        obj: Pascal cased object

    Returns:
        Snake cased object
    """
    working: list[Any] | dict[Any, Any]
    if isinstance(obj, list):
        working = [pascal_to_snake(x) for x in obj]
        return working
    if isinstance(obj, dict):
        working = {}
        for k, val in obj.items():
            new_key = PASCAL_REGEX.sub(r"_\1", k).lower()
            working[new_key] = pascal_to_snake(val)
        return working
    return obj


def path_is_relative_to(child: Path, parent: Path) -> bool:
    """Return True if the path is relative to another path or False.

    Args:
        child: The path that may be a child
        parent: The path that may be a parent

    Returns:
        Indicates the child is a child of the parent
    """
    return child.is_relative_to(parent)


def remove_ansi(string: str) -> str:
    """Strip ansi code from a str.

    Args:
        string: String to strip ansi code from

    Returns:
        String without ansi code
    """
    ansi_escape = re.compile(
        r"""
            \x1B  # ESC
            (?:   # 7-bit C1 Fe (except CSI)
                [@-Z\\-_]
            |     # or [ for CSI, followed by a control sequence
                \[
                [0-?]*  # Parameter bytes
                [ -/]*  # Intermediate bytes
                [@-~]   # Final byte
            )
        """,
        re.VERBOSE,
    )
    return ansi_escape.sub("", string)


def remove_dbl_un(string: str) -> str:
    """Remove a __ from the beginning of a string.

    Args:
        string: String to remove __ from

    Returns:
        String without __
    """
    if string.startswith("__"):
        return string.replace("__", "", 1)
    return string


def round_half_up(number: float) -> int:
    """Round a number to the nearest integer with ties going away from zero.

    This is different the round() where exact halfway cases are rounded to the nearest
    even result instead of away from zero. (e.g. round(2.5) = 2, round(3.5) = 4).

    This will always round based on distance from zero. (e.g round(2.5) = 3, round(3.5) = 4).

    Args:
        number: The number to round

    Returns:
        The rounded number as an it
    """
    rounded = decimal.Decimal(number).quantize(decimal.Decimal(1), rounding=decimal.ROUND_HALF_UP)
    return int(rounded)


def shlex_join(tokens: Iterable[str]) -> str:
    """Concatenate the tokens of a list and return a string.

    Args:
        tokens: The iterable of strings to join

    Returns:
        The iterable joined with spaces
    """
    return shlex.join(split_command=tokens)


def str2bool(value: Any) -> bool:
    """Convert some commonly used values to a boolean.

    Args:
        value: Value to convert to boolean

    Raises:
        ValueError: If value is not a boolean or string

    Returns:
        New converted boolean
    """
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        if value.lower() in ("yes", "true"):
            return True
        if value.lower() in ("no", "false"):
            return False
    raise ValueError


# We are kind-of screwed type-wise by the fact that ast.literal_eval()
#  returns Any. Need to find a better solution... "Any" isn't it.
def templar(string: str, template_vars: Mapping[Any, Any]) -> tuple[list[str], Any]:
    """Template some string with jinja2 always to and from json.

    Args:
        string: The template string
        template_vars: The vars used to render the template

    Returns:
        A list of errors and either the result of templating or original
        string
    """
    errors = []
    # hide the jinja that may be in the template_vars
    template_vars = escape_moustaches(template_vars)

    env = Environment(autoescape=True, undefined=StrictUndefined)
    try:
        template: Any = env.from_string(string)
        result = template.render(template_vars)
    except (ValueError, TemplateError) as exc:
        errors.append(f"Error while templating string: '{string}'")
        errors.append(f"The error was: {exc!s}")
        for error in errors:
            logger.exception(error)
        return errors, string

    # We may have gotten the __repr__ of a python object
    # so let's try and turn it back
    try:
        logger.debug("original templated string: %s", result)
        escaped = html.unescape(result)
        logger.debug("html escaped templated str: %s", escaped)
        result = ast.literal_eval(escaped)
    except (ValueError, SyntaxError) as exc:
        logger.debug("Could not ast parse templated string")
        logger.debug("error was: %s", str(exc))
        logger.debug("attempted on %s", result)

    result = unescape_moustaches(result)  # type: ignore[assignment, arg-type]
    return errors, result


def timestamp_to_iso(timestamp: float, time_zone: str) -> str | None:
    """Generate an ISO 8601 date time string from a timestamp.

    Args:
        timestamp: The unix timestamp
        time_zone: The time zone

    Returns:
        The ISO string
    """
    try:
        if time_zone == "local":
            return (
                datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)
                .astimezone()
                .isoformat()
            )
        if time_zone == "UTC":
            return datetime.datetime.fromtimestamp(
                timestamp,
                tz=datetime.timezone.utc,
            ).isoformat()
        return datetime.datetime.fromtimestamp(
            timestamp,
            tz=zoneinfo.ZoneInfo(time_zone),
        ).isoformat()
    except zoneinfo.ZoneInfoNotFoundError:
        logger.exception("The time zone '%s' could not be found. Returning None", time_zone)
        return None


def time_stamp_for_file(path: str, time_zone: str) -> tuple[float | None, str | None]:
    """Get a timestamp for a file path.

    Args:
        path: The file path
        time_zone: Time zone

    Returns:
        The UNIX timestamp and an ISO 8601 string
    """
    try:
        modified = Path(path).stat().st_mtime
    except FileNotFoundError:
        # It may have been mounted to a different location in the execution environment
        modified = None
    if modified is not None:
        iso_stamp = timestamp_to_iso(timestamp=modified, time_zone=time_zone)
    else:
        iso_stamp = None
    return modified, iso_stamp


def to_list(thing: str | list[Any] | tuple[Any] | set[Any] | None) -> list[Any]:
    """Convert something to a list if necessary.

    Args:
        thing: Item to convert to a list

    Returns:
        Item as a list
    """
    if isinstance(thing, list | tuple | set):
        converted_value = list(thing)
    elif thing is not None:
        converted_value = [thing]
    else:
        converted_value = []
    return converted_value


def unescape_moustaches(obj: Any) -> Any:
    """Unescape moustaches.

    Args:
        obj: Variable that needs to contain moustaches

    Returns:
        The obj with replacements made
    """
    replacements = (("U+007B", "{"), ("U+007D", "}"))
    result = dispatch(obj, replacements)
    return result



"""Config subcommand implementation."""

from __future__ import annotations

import curses
import re
import shlex
import shutil

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.runner import AnsibleConfig
from ansible_navigator.runner import Command
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import nonblocking_notification
from ansible_navigator.ui_framework import warning_notification
from ansible_navigator.utils.serialize import Loader
from ansible_navigator.utils.serialize import yaml

from . import _actions as actions
from . import run_action


if TYPE_CHECKING:
    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration


def color_menu(colno: int, colname: str, entry: dict[str, Any]) -> tuple[int, int]:
    """Provide a color for a collections menu entry in one column.

    Args:
        colno: The column number
        colname: The column name
        entry: The menu entry

    Returns:
        The color and decoration
    """
    if entry["default"] is False:
        return 3, 0
    return 2, 0


def content_heading(obj: Any, screen_w: int) -> CursesLines | None:
    """Create a heading for config content.

    Args:
        obj: The content going to be shown
        screen_w: The current screen width

    Returns:
        The heading
    """
    string = obj["name"]
    if obj["default"] is False:
        string += f" (current: {obj['current_value']})  (default: {obj['default_value']})"
        color = 3
    else:
        string += f" (current/default: {obj['current_value']})"
        color = 2

    string = string + (" " * (screen_w - len(string) + 1))

    line_part = CursesLinePart(
        column=0,
        string=string,
        color=color,
        decoration=curses.A_UNDERLINE,
    )
    return CursesLines((CursesLine((line_part,)),))


def filter_content_keys(obj: dict[Any, Any]) -> dict[Any, Any]:
    """Filter out some keys when showing collection content.

    Args:
        obj: The object from which keys should be removed

    Returns:
        The object with keys removed
    """
    return {k: v for k, v in obj.items() if not k.startswith("__")}


@actions.register
class Action(ActionBase):
    """Config subcommand implementation."""

    KEGEX = r"^config(\s(?P<params>.*))?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:config`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="config")

        self._config: list[Any] | None = None
        self._runner: AnsibleConfig | Command

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the ``config`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("config requested in interactive mode")
        self._prepare_to_run(app, interaction)

        notification = nonblocking_notification(
            messages=[
                "Collecting the ansible configuration, this may take a minute...",
            ],
        )
        interaction.ui.show_form(notification)

        args_updated = self._update_args(
            [self._name] + shlex.split(self._interaction.action.match.groupdict()["params"] or ""),
        )
        if not args_updated:
            self._prepare_to_exit(interaction)
            return None

        self._run_runner()
        if self._config is None:
            self._prepare_to_exit(interaction)
            return None

        self.steps.append(self._build_main_menu())

        while True:
            self._calling_app.update()
            self._take_step()

            if not self.steps:
                break

            if self.steps.current.name == "quit":
                return self.steps.current

        self._prepare_to_exit(interaction)
        return None

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``config`` request for mode stdout.

        Returns:
            The return code or 1. If the response from the runner
            invocation is None, indicates there is no console output to
            display, so assume an issue and return 1 along with a
            message to review the logs.
        """
        self._logger.debug("config requested in stdout mode")
        response = self._run_runner()
        if (
            response is None
            or response[0] is None
            or response[1] is None
            or not isinstance(response[2], int)
        ):
            self._logger.error("Unexpected response: %s", response)
            return RunStdoutReturn(message="Please review the log for errors.", return_code=1)
        _out, error, return_code = response
        return RunStdoutReturn(message=error, return_code=return_code)

    def _take_step(self) -> None:
        """Take a step based on the current step or step back."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, Step):
            if self.steps.current.show_func:
                current_index = self.steps.current.index
                self.steps.current.show_func()
                self.steps.current.index = current_index

            if self.steps.current.type == "menu":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    columns=self.steps.current.columns,
                    color_menu_item=color_menu,
                )
            elif self.steps.current.type == "content":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    index=self.steps.current.index,
                    content_heading=content_heading,
                    filter_content_keys=filter_content_keys,
                )

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)

    def _build_main_menu(self) -> Step:
        """Build the menu of configuration options.

        Returns:
            The config menu definition
        """
        return Step(
            name="all_options",
            columns=["name", "default", "source", "__current"],
            select_func=self._build_option_content,
            step_type="menu",
            value=self._config,
        )

    def _build_option_content(self) -> Step:
        """Build the content for one configuration option.

        Returns:
            The option's content
        """
        return Step(
            name="option_content",
            step_type="content",
            value=self._config,
            index=self.steps.current.index,
        )

    def _run_runner(self) -> tuple[str, str, int] | None:
        """Use the runner subsystem to retrieve the configuration.

        Returns:
            For mode interactive nothing. For mode stdout the output,
            errors and return code from runner.
        """
        if isinstance(self._args.set_environment_variable, dict):
            set_env_vars = {**self._args.set_environment_variable}
        else:
            set_env_vars = {}

        if self._args.display_color is False:
            set_env_vars["ANSIBLE_NOCOLOR"] = "1"

        kwargs = {
            "container_engine": self._args.container_engine,
            "host_cwd": str(Path.cwd()),
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "navigator_mode": self._args.mode,
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": set_env_vars,
            "private_data_dir": self._args.ansible_runner_artifact_dir,
            "rotate_artifacts": self._args.ansible_runner_rotate_artifacts_count,
            "timeout": self._args.ansible_runner_timeout,
        }

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs.update(
                {"container_volume_mounts": self._args.execution_environment_volume_mounts},
            )

        if isinstance(self._args.container_options, list):
            kwargs.update({"container_options": self._args.container_options})

        if self._args.mode == "interactive":
            self._run_runner_interactive(kwargs)
            return None
        return self._run_runner_stdout(kwargs)

    def _run_runner_interactive(
        self,
        kwargs: dict[str, Any],
    ) -> None:
        """Run the ansible-config runner in interactive mode.

        Fetches both the list and dump output from ansible-config,
        logs any errors, and parses/merges the results.

        Args:
            kwargs: The common runner keyword arguments.
        """
        self._runner = AnsibleConfig(**kwargs)
        config_file = self._args.config if isinstance(self._args.config, str) else None
        list_output, list_output_err = self._runner.fetch_ansible_config(
            "list",
            config_file=config_file,
        )
        dump_output, dump_output_err = self._runner.fetch_ansible_config(
            "dump",
            config_file=config_file,
        )
        if list_output_err:
            msg = f"Error occurred while fetching ansible config (list): '{list_output_err}'"
            self._logger.error(msg)
        if dump_output_err:
            msg = f"Error occurred while fetching ansible config (dump): '{dump_output_err}'"
            self._logger.error(msg)

        err_msg = "\n".join(e for e in (list_output_err, dump_output_err) if e)
        if "ERROR!" in err_msg or not list_output or not dump_output:
            warn_msg = ["Errors were encountered while gathering the configuration:"]
            if err_msg:
                warn_msg += err_msg.splitlines()
            if not list_output or not dump_output:
                warn_msg.append("The configuration could not be gathered.")
            warning = warning_notification(warn_msg)
            self._interaction.ui.show_form(warning)
        else:
            self._parse_and_merge(list_output, dump_output)

    def _run_runner_stdout(
        self,
        kwargs: dict[str, Any],
    ) -> tuple[str, str, int] | None:
        """Run the ansible-config runner in stdout mode.

        Resolves the ansible-config executable path, builds the
        pass-through arguments, and runs the command.

        Args:
            kwargs: The common runner keyword arguments.

        Raises:
            RuntimeError: When the ansible-config command cannot be
                found with execution environment support disabled.

        Returns:
            The output, errors and return code from runner.
        """
        if self._args.execution_environment:
            ansible_config_path = "ansible-config"
        else:
            exec_path = shutil.which("ansible-config")
            if exec_path is None:
                msg = "'ansible-config' executable not found"
                self._logger.error(msg)
                raise RuntimeError(msg)
            ansible_config_path = exec_path

        pass_through_arg = self._args.cmdline.copy() if isinstance(self._args.cmdline, list) else []

        if self._args.help_config is True:
            pass_through_arg.append("--help")

        if isinstance(self._args.config, str):
            pass_through_arg.extend(["--config", self._args.config])

        kwargs.update({"cmdline": pass_through_arg})

        self._runner = Command(executable_cmd=ansible_config_path, **kwargs)
        stdout_return = self._runner.run()
        return stdout_return

    def _parse_and_merge(self, list_output: str, dump_output: str) -> None:
        """Parse the list and dump output. Merge dump into list.

        Args:
            list_output: The output from config list
            dump_output: The output from config dump

        Returns:
            Nothing
        """
        try:
            parsed = yaml.load(list_output, Loader=Loader)
            self._logger.debug("yaml loading list output succeeded")
        except yaml.YAMLError as exc:
            self._logger.debug("error yaml loading list output: '%s'", str(exc))
            return

        parsed.pop("GALAXY_SERVERS", None)

        regex = re.compile(r"^(?P<variable>\S+)\((?P<source>.*)\)\s=\s(?P<current>.*)$")
        for line in dump_output.splitlines():
            if not line:
                break
            if not self._parse_dump_line(regex, line, parsed):
                return

        self._finalize_parsed_config(parsed)
        self._config = list(parsed.values())
        self._logger.debug("parsed and merged list and dump successfully")
        return

    def _parse_dump_line(
        self,
        regex: re.Pattern[str],
        line: str,
        parsed: dict[str, Any],
    ) -> bool:
        """Parse a single line from the ansible-config dump output.

        Extracts the variable name, source, and current value from
        the line and merges them into the parsed config dictionary.

        Args:
            regex: The compiled regex pattern for matching dump lines.
            line: A single line from the dump output.
            parsed: The parsed config dictionary to update in place.

        Returns:
            True if the line was parsed successfully, False on error.
        """
        extracted = regex.match(line)
        if not extracted:
            self._logger.error("Unparsable dump entry: %s", line)
            return False

        variable = extracted.groupdict()["variable"]
        try:
            source = yaml.load(extracted.groupdict()["source"], Loader=Loader)
        except yaml.YAMLError:
            source = extracted.groupdict()["source"]
        try:
            current = yaml.load(extracted.groupdict()["current"], Loader=Loader)
        except yaml.YAMLError:
            current = extracted.groupdict()["current"]
        try:
            if isinstance(source, dict):
                parsed[variable]["source"] = next(iter(source.keys()))
                parsed[variable]["via"] = next(iter(source.values()))
            else:
                parsed[variable]["source"] = source
                parsed[variable]["via"] = source
            current_as_str = str(current)

            target_screen_w = int(100 / 2)  # half a wide screen
            if len(current_as_str) > target_screen_w:
                more_indicator = "..."
                text_width = target_screen_w - len(more_indicator)
                current_as_str = f"{current_as_str[0:text_width]}{more_indicator}"
            parsed[variable]["__current"] = current_as_str
            parsed[variable]["current_value"] = current
        except KeyError:
            self._logger.exception("variable '%s' not found in list output", variable)
            return False
        return True

    def _finalize_parsed_config(self, parsed: dict[str, Any]) -> None:
        """Post-process the parsed config entries with display metadata.

        Adds option name, default indicators, and config file path
        to each parsed config entry.

        Args:
            parsed: The parsed config dictionary to update in place.
        """
        for key, value in parsed.items():
            value["option"] = key
            value["name"] = key.replace("_", " ").capitalize()
            value["default_value"] = value.get("default", "None")
            if value["source"] == "default":
                value["default"] = True
            else:
                value["default"] = False
            value["current_config_file"] = str(self._args.internals.ansible_configuration.path)

        self._config = list(parsed.values())
        self._logger.debug("parsed and merged list and dump successfully")



from __future__ import annotations

import functools
import re

from re import Match
from typing import TYPE_CHECKING

import onigurumacffi

from .region import Region


if TYPE_CHECKING:
    from ._types import Protocol  # noqa: F401
    from .compiler import Compiler
    from .region import Regions
    from .rules import CompiledRegsetRule
    from .rules import CompiledRule  # noqa: F401
    from .rules import _Rule  # noqa: F401
    from .state import State

_BACKREF_RE = re.compile(r"((?<!\\)(?:\\\\)*)\\(\d+)")


_FLAGS = {
    # (first_line, boundary)
    (False, False): (
        onigurumacffi.OnigSearchOption.NOT_END_STRING
        | onigurumacffi.OnigSearchOption.NOT_BEGIN_STRING
        | onigurumacffi.OnigSearchOption.NOT_BEGIN_POSITION
    ),
    (False, True): (
        onigurumacffi.OnigSearchOption.NOT_END_STRING
        | onigurumacffi.OnigSearchOption.NOT_BEGIN_STRING
    ),
    (True, False): (
        onigurumacffi.OnigSearchOption.NOT_END_STRING
        | onigurumacffi.OnigSearchOption.NOT_BEGIN_POSITION
    ),
    (True, True): onigurumacffi.OnigSearchOption.NOT_END_STRING,
}


class _Reg:
    def __init__(self, s: str) -> None:
        self._pattern = s
        self._reg = onigurumacffi.compile(self._pattern)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._pattern!r})"

    def search(self, line: str, pos: int, first_line: bool, boundary: bool) -> Match[str] | None:
        return self._reg.search(line, pos, flags=_FLAGS[first_line, boundary])

    def match(self, line: str, pos: int, first_line: bool, boundary: bool) -> Match[str] | None:
        return self._reg.match(line, pos, flags=_FLAGS[first_line, boundary])


class _RegSet:
    def __init__(self, *s: str) -> None:
        self._patterns = s
        self._set = onigurumacffi.compile_regset(*self._patterns)

    def __repr__(self) -> str:
        args = ", ".join(repr(s) for s in self._patterns)
        return f"{type(self).__name__}({args})"

    def search(
        self,
        line: str,
        pos: int,
        first_line: bool,
        boundary: bool,
    ) -> tuple[int, Match[str] | None]:
        return self._set.search(line, pos, flags=_FLAGS[first_line, boundary])


def do_regset(
    idx: int,
    match: Match[str] | None,
    rule: CompiledRegsetRule,
    compiler: Compiler,
    state: State,
    pos: int,
) -> tuple[State, int, bool, Regions] | None:
    if match is None:
        return None

    ret = []
    if match.start() > pos:
        ret.append(Region(pos, match.start(), state.cur.scope))

    target_rule = compiler.compile_rule(rule.u_rules[idx])
    state, boundary, regions = target_rule.start(compiler, match, state)
    ret.extend(regions)

    return state, match.end(), boundary, tuple(ret)


def expand_escaped(match: Match[str], s: str) -> str:
    return _BACKREF_RE.sub(lambda m: f"{m[1]}{re.escape(match[int(m[2])])}", s)


make_reg = functools.cache(_Reg)
make_regset = functools.cache(_RegSet)
ERR_REG = make_reg("$ ^")



"""Builder subcommand implementation.

Importing this module registers this subcommand in the external
global subcommand registry.
"""

from __future__ import annotations

import shutil

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.configuration_subsystem.definitions import Constants
from ansible_navigator.runner import Command
from ansible_navigator.utils.functions import expand_path

from . import _actions as actions


@actions.register
class Action(ActionBase):
    """Run the builder subcommand."""

    KEGEX = "^b(?:uilder)?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the action.

        Args:
            args: The current application configuration
        """
        super().__init__(args=args, logger_name=__name__, name="builder")

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``builder`` request for mode stdout.

        Returns:
            The return code or 1. If the response from the runner
            invocation is None, indicates there is no console output to
            display, so assume an issue and return 1 along with a
            message to review the logs.
        """
        self._logger.debug("builder requested in stdout mode")
        response = self._run_runner()
        if response is None:
            self._logger.error("Unexpected response: %s", response)
            return RunStdoutReturn(message="Please review the log for errors.", return_code=1)
        _out, error, return_code = response
        return RunStdoutReturn(message=error, return_code=return_code)

    def _run_runner(self) -> tuple[str, str, int] | None:
        """Spin up runner.

        Raises:
            RuntimeError: When ansible-builder can not be found

        Returns:
            The stdout, stderr and return code from runner
        """
        ansible_builder_path = shutil.which("ansible-builder")
        if ansible_builder_path is None:
            msg = "'ansible-builder' executable not found"
            self._logger.error(msg)
            raise RuntimeError(msg)

        if isinstance(self._args.set_environment_variable, dict):
            env_vars_to_set = self._args.set_environment_variable.copy()
        elif isinstance(self._args.set_environment_variable, Constants):
            env_vars_to_set = {}
        else:
            log_message = (
                "The setting 'set_environment_variable' was neither a dictionary"
                " or Constants, please raise an issue. No environment variables will be set."
            )
            self._logger.error(
                "%s The current value was found to be '%s'",
                log_message,
                self._args.set_environment_variable,
            )
            env_vars_to_set = {}

        if self._args.display_color is False:
            env_vars_to_set["ANSIBLE_NOCOLOR"] = "1"

        if self._args.execution_environment:
            self._logger.info("For builder subcommand execution-environment is disabled")

        kwargs = {
            "execution_environment": False,
            "host_cwd": str(expand_path(self._args.workdir)),
            "navigator_mode": self._args.mode,
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": env_vars_to_set,
            "timeout": self._args.ansible_runner_timeout,
        }

        pass_through_arg = []

        if isinstance(self._args.cmdline, list):
            pass_through_arg.extend(self._args.cmdline)

        if self._args.help_builder is True:
            pass_through_arg.append("--help")

        kwargs.update({"cmdline": pass_through_arg})

        command_runner = Command(executable_cmd=ansible_builder_path, **kwargs)
        return command_runner.run()



"""A mechanism to start the main application loop for the text user interface.

An instance of :class:`~ansible_navigator.action_runner.ActionRunner`
is invoked once, after settings are parsed. A loop is entered and only
exited when the user requests to quit.

After initializing the user interface, an :class:`~ansible_navigator.ui_framework.ui.Interaction`
is created from the fully initialized settings, and the requested
action is run.

From that point forward flow control of the application is handled
by each action in the action stack, control returned here when
``:quit`` is requested by the user.
"""

from typing import TYPE_CHECKING

from ansible_navigator.actions import kegexes
from ansible_navigator.actions import run_action

from .action_base import ActionBase
from .configuration_subsystem.definitions import ApplicationConfiguration
from .constants import GRAMMAR_DIR
from .constants import TERMINAL_COLORS_PATH
from .constants import THEME_PATH
from .steps import Steps
from .ui_framework import Interaction
from .ui_framework import UIConfig
from .ui_framework import UserInterface


if TYPE_CHECKING:
    from _curses import window

    Window = window
else:
    from typing import Any

    Window = Any

DEFAULT_REFRESH = 100


class ActionRunner(ActionBase):
    """A single action runner."""

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ActionRunner class.

        Args:
            args: The current application configuration
        """
        super().__init__(args, name="action_runner")
        self._ui: UserInterface
        self.steps: Steps = Steps()

    def initialize_ui(self, refresh: int) -> None:
        """Initialize the user interface.

        Args:
            refresh (int): The refresh for the UI
        """
        config = UIConfig(
            color=self._args.display_color,
            colors_initialized=False,
            grammar_dir=GRAMMAR_DIR,
            osc4=self._args.osc4,
            terminal_colors_path=TERMINAL_COLORS_PATH,
            theme_path=THEME_PATH,
        )
        self._logger.debug("grammar path = %s", config.grammar_dir)
        self._logger.debug("theme path = %s", config.theme_path)
        self._logger.debug("terminal colors path = %s", config.terminal_colors_path)

        self._ui = UserInterface(
            screen_min_height=3,
            kegexes=kegexes,
            refresh=refresh,
            ui_config=config,
        )

    def run(self, _screen: Window) -> None:
        """Run the app.

        Initialize the UI.
        Create an interaction based on the app name from the current settings.
        Run the action, passing the interaction.

        Args:
            _screen: The screen instance from the curses wrapper call
        """
        self.initialize_ui(DEFAULT_REFRESH)
        name, action = self._action_match(self._args.app)
        if name and action:
            interaction = Interaction(
                name=name,
                action=action,
                menu=None,
                content=None,
                ui=self._ui._ui,  # noqa: SLF001
            )
            self._run_app(interaction)

    def _run_app(self, initial_interaction: Interaction) -> None:
        """Enter the endless app loop.

        Args:
            initial_interaction: The initial interaction for app start
        """
        interaction = None
        while True:
            if not self.steps:
                self.steps.append(initial_interaction)

            if isinstance(self.steps.current, Interaction):
                interaction = run_action(self.steps.current.name, self.app, self.steps.current)
            if interaction is None:
                self.steps.back_one()
                if not self.steps:
                    break
            elif interaction.name == "quit":
                break
            else:
                self.steps.append(interaction)



"""Predefined form field validators."""

from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from random import randrange
from typing import TYPE_CHECKING
from typing import Any
from typing import NamedTuple
from urllib.parse import urlparse

from ansible_navigator.utils.functions import expand_path

from .sentinels import Unknown
from .sentinels import unknown


if TYPE_CHECKING:
    from .form_defs import FieldValidationStates


class Validation(NamedTuple):
    """The response from a validation."""

    value: Any
    error_msg: str


class FieldValidators:
    """A box in which field validators are put."""

    @staticmethod
    def http(text: str = "", hint: bool = False) -> Validation | str:
        """Validate http or https.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = "Please enter a valid URL"
        if hint:
            return msg
        if text:
            result = urlparse(text)
            if result.scheme in ["http", "https"] and result.hostname:
                msg = ""
        return Validation(value=text, error_msg=msg)

    @staticmethod
    def masked_or_none(text: str = "", hint: bool = False) -> Validation | str:
        """Validate a masked field or None.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        if hint:
            return "Please provide a value (optional)"
        value = "*" * randrange(15, 20) if text else ""  # noqa:S311
        return Validation(value=value, error_msg="")

    @staticmethod
    def none(
        text: FieldValidationStates | str = "",
        hint: bool = False,
    ) -> Validation | str:
        """Validate nothing about this field.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        if hint:
            return "Please provide a value (optional)"
        return Validation(value=text, error_msg="")

    @staticmethod
    def null(text: str = "", hint: bool = False) -> Validation | str:
        """No validation, no message.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        if hint:
            return ""
        return Validation(value=text, error_msg="")

    @staticmethod
    def one_of(choices: list[str] = [], text: str = "", hint: bool = False) -> Validation | str:
        """Validate that some text is one of choices.

        Args:
            choices: The list of choices
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        choices_str = f"{', '.join(choices[:-1])} or {choices[-1]}" if choices else ""
        msg = f"Please enter {choices_str}"
        value = text
        if hint:
            return msg
        try:
            value = choices[[c.lower() for c in choices].index(text.lower())]
            msg = ""
        except ValueError:
            pass
        return Validation(value=value, error_msg=msg)

    @staticmethod
    def some_of_or_none(
        choices: Unknown | list[Any] = unknown,
        min_selected: Unknown | int = unknown,
        max_selected: Unknown | int = unknown,
        hint: bool = False,
    ) -> Validation | str:
        """Validate a checkbox set.

        Args:
            choices: The list of choices
            min_selected: The minimum number of choices that must be
                selected
            max_selected: The maximum number of choices that can be
                selected
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message

        Raises:
            TypeError: If the types are not as expected
        """
        if min_selected == max_selected:
            word = "entry" if min_selected == 1 else "entries"
            msg = f"Please select {min_selected!s} {word}"
        else:
            msg = f"Please select between {min_selected} and "
            word = str(max_selected) if max_selected != -1 else "all"
            msg += f"{word} entries"

        if hint:
            return msg
        if (
            isinstance(min_selected, int)
            and isinstance(max_selected, int)
            and isinstance(choices, Iterable)
        ):
            checked = len([choice for choice in choices if choice.checked])
            if max_selected >= checked >= min_selected:
                msg = ""

            return Validation(value=choices, error_msg=msg)
        raise TypeError

    @staticmethod
    def something(text: str = "", hint: bool = False) -> Validation | str:
        """Validate that the text is not an empty string.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = "Please provide a value (required)"
        if hint:
            return msg
        if text:
            msg = ""
        return Validation(value=text, error_msg=msg)

    @staticmethod
    def true_false(text: str = "", hint: bool = False) -> Validation | str:
        """Validate true or false.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = "Please enter true or false"
        if hint:
            return msg
        if text:
            if text[0].lower() == "t":
                value = True
                msg = ""
                return Validation(value=value, error_msg=msg)

            if text[0].lower() == "f":
                value = False
                msg = ""
                return Validation(value=value, error_msg=msg)
        return Validation(value=text, error_msg=msg)

    @staticmethod
    def valid_file_path(text: str = "", hint: bool = False) -> Validation | str:
        """Validate that a file path is a real file.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = "Please enter a valid file path"
        if hint:
            return msg
        value = str(expand_path(text))
        if Path(value).exists() and Path(value).is_file():
            msg = ""
        else:
            value = text
        return Validation(value=value, error_msg=msg)

    @staticmethod
    def valid_path(text: str = "", hint: bool = False) -> Validation | str:
        """Validate that a path is real.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = "Please enter a valid file or directory path"
        if hint:
            return msg
        value = str(expand_path(text))
        if Path(value).exists():
            msg = ""
        else:
            value = text
        return Validation(value=value, error_msg=msg)

    @staticmethod
    def valid_path_or_none(text: str = "", hint: bool = False) -> Validation | str:
        """Validate that a path is real or none.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = "Please enter a valid path or leave blank"
        if hint:
            return msg

        if text == "":
            msg = ""
            value = text
        else:
            value = str(expand_path(text))
            if Path(value).exists():
                msg = ""
            else:
                value = text
        return Validation(value=value, error_msg=msg)

    @staticmethod
    def yes_no(text: str = "", hint: bool = False) -> Validation | str:
        """Validate yes or no.

        Args:
            text: The text to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = "Please enter yes or no"
        value = text
        if hint:
            return msg
        if text:
            if text[0] == "y":
                value = "yes"
                msg = ""
            elif text[0] == "n":
                value = "no"
                msg = ""
        return Validation(value=value, error_msg=msg)


class FormValidators:
    """Validators for a form."""

    @staticmethod
    def all_true(response: list[Any] | None = None, hint: bool = False) -> Validation | str:
        """Validate all in list are true.

        Args:
            response: The list to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = "Please ensure all values are true"
        if hint:
            return msg
        response = response or []
        if all(v is True for v in response):
            msg = ""
        return Validation(value=response, error_msg=msg)

    @staticmethod
    def no_validation(
        response: list[Any] | None = None,
        hint: bool = False,
    ) -> Validation | str:
        """No validation.

        Args:
            response: The list to validate
            hint: If True, return a hint message instead of a Validation

        Returns:
            A Validation or a hint message
        """
        msg = ""
        if hint:
            return msg
        return Validation(value=response, error_msg=msg)



"""Image puller."""

from __future__ import annotations

import logging
import os
import subprocess

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.utils.definitions import ExitMessage
from ansible_navigator.utils.definitions import ExitPrefix
from ansible_navigator.utils.definitions import LogMessage
from ansible_navigator.utils.functions import shlex_join


if TYPE_CHECKING:
    from ansible_navigator.configuration_subsystem import Constants


@dataclass(frozen=False)
class ImageAssessment:
    """Data structure containing the image assessment.

    An ``ImageAssessment`` gets updated after instantiation
    with the determination of whether or not a pull is required.
    """

    messages: list[LogMessage]
    exit_messages: list[ExitMessage]
    pull_required: bool


class ImagePuller:
    # pylint: disable=too-many-instance-attributes
    """Image puller."""

    def __init__(
        self,
        container_engine: str,
        image: str,
        arguments: Constants | list[str],
        pull_policy: str,
    ) -> None:
        """Initialize the container image puller.

        Args:
            container_engine: The name of the container engine
            image: The name of the image to pull
            arguments: Additional arguments to be appended to the pull
                policy
            pull_policy: The current pull policy from the settings
        """
        if isinstance(arguments, list):
            self._arguments = arguments
        else:
            self._arguments = []

        self._assessment = ImageAssessment(messages=[], exit_messages=[], pull_required=False)
        self._container_engine: str = container_engine
        self._exit_messages: list[ExitMessage] = []
        self._image: str = image
        self._image_present: bool
        self._image_tag: str
        self._logger = logging.getLogger(__name__)
        self._messages: list[LogMessage] = []
        self._pull_policy: str = pull_policy
        self._pull_required: bool = False

    def assess(self) -> None:
        """Assess the need to pull."""
        self._extract_tag()
        self._check_for_image()
        self._determine_pull()
        if self._pull_policy == "never" and self._image_present is False:
            exit_msg = (
                "Pull policy is set to 'never' and execution environment"
                " image was not found locally"
            )
            self._log_message(message=exit_msg, level=logging.ERROR)
            exit_msg = (
                "Try again with '--pp missing' or manually pull the execution environment image"
            )
            self._log_message(level=logging.INFO, message=exit_msg, hint=True)

        self._assessment = ImageAssessment(
            messages=self._messages,
            exit_messages=self._exit_messages,
            pull_required=self._pull_required,
        )

    @property
    def assessment(self) -> ImageAssessment:
        """Return an image assessment.

        Returns:
            Image assessment
        """
        return self._assessment

    def _check_for_image(self) -> None:
        """Check for the image."""
        try:
            cmd_parts = [self._container_engine, "image", "inspect", self._image]
            self._log_message(level=logging.DEBUG, message=f"Command: {shlex_join(cmd_parts)}")
            subprocess.run(
                cmd_parts,
                check=True,
                capture_output=True,
            )
            self._image_present = True

        except subprocess.CalledProcessError as exc:
            self._image_present = False
            stdout = exc.stdout.decode()
            stderr = exc.stderr.decode()
            self._log_message(level=logging.DEBUG, message=f"stdout: {stdout}")
            self._log_message(level=logging.DEBUG, message=f"stderr: {stderr}")
            if "no such image" not in str(exc.stderr).lower():
                message = "Image inspection failed, image assumed to be corrupted or missing"
                self._log_message(level=logging.WARNING, message=message)
                self._log_message(level=logging.WARNING, message=f"stdout: {stdout}")
                self._log_message(level=logging.WARNING, message=f"stderr: {stderr}")

    def _determine_pull(self) -> None:
        """Determine if a pull is required."""
        if self._pull_policy == "missing" and self._image_present is False:  # noqa: SIM114
            pull = True
        elif self._pull_policy == "always":  # noqa: SIM114
            pull = True
        elif self._pull_policy == "tag" and self._image_tag == "latest":  # noqa: SIM114
            pull = True
        elif self._pull_policy == "tag" and self._image_present is False:
            pull = True
        else:
            pull = False
        self._pull_required = pull
        self._assessment.pull_required = pull

    def _extract_tag(self) -> None:
        """Extract the tag from the image name."""
        image_pieces = self._image.split(":")
        self._image_tag = "latest"
        if len(image_pieces) > 1:
            tag = image_pieces[-1]
            if "/" not in tag:
                # If there's a slash in the tag, it means we were probably given
                # an image name that has a port number but no tag at the end.
                # e.g.: registry.example.com:443/my/image
                # In this case, we /don't/ want "443/my/image" to be considered
                # a tag. Only if the ending doesn't have a slash in it, is it a
                # valid tag.
                self._image_tag = tag
        message = f"Image tag is: {self._image_tag}"
        self._log_message(level=logging.INFO, message=message)

    def _log_message(self, message: str, level: int, hint: bool = False) -> None:
        """Log a message.

        Args:
            message: The message to log
            level: The logging level
            hint: Whether or not the message is a hint
        """
        if level == logging.ERROR:
            self._exit_messages.append(ExitMessage(message=message))
        elif hint:
            self._exit_messages.append(ExitMessage(message=message, prefix=ExitPrefix.HINT))
        else:
            self._messages.append(LogMessage(level=level, message=message))
        self._logger.log(level=level, msg=message)

    def prologue_stdout(self) -> None:
        """Print a little value added information about the execution environment."""
        messages: list[tuple[str, Any]] = [("Execution environment image name:", self._image)]
        messages.append(("Execution environment image tag:", self._image_tag))
        arguments = shlex_join(self._arguments) or None
        messages.append(("Execution environment pull arguments:", arguments))
        messages.append(("Execution environment pull policy:", self._pull_policy))
        messages.append(("Execution environment pull needed:", self._pull_required))

        width = max(len(m[0]) + len(str(m[1])) + 2 for m in messages)
        print("\u002d" * width)
        print("Execution environment image and pull policy overview")
        print("\u002d" * width)
        column_width = max(len(m[0]) for m in messages)
        for msg, value in messages:
            print(f"{msg.ljust(column_width)} {value!s}")
            self._log_message(message=f"{msg!s}: {value!s}", level=logging.INFO)
        print("\u002d" * width)
        print("Updating the execution environment")
        print("\u002d" * width)

    def _generate_pull_command(self) -> str:
        """Generate the pull command.

        Returns:
            The command
        """
        command_line = [self._container_engine, "pull"]
        command_line.extend(self._arguments)
        command_line.append(self._image)
        joined_command = " ".join(command_line)
        return joined_command

    def pull_stdout(self) -> None:
        """Pull the image, print to stdout.

        ``podman`` writes errors to stdout and ``docker`` writes to stderr.

        This allows us to capture, log and color the error from ``docker`` but in the case
        of ``podman`` the error is only printed on the screen.

        In both cases, stdout is not captured so the user can see the progress on the screen
        as the pull is happening.
        """
        try:
            command_line = self._generate_pull_command()
            cmd_to_run = f"echo Running the command: {command_line} && {command_line}"
            stderr_pipe = subprocess.PIPE if self._container_engine == "docker" else None
            subprocess.run(
                cmd_to_run,
                check=True,
                stderr=stderr_pipe,
                shell=True,
                env=os.environ,
            )
            self._log_message(level=logging.INFO, message="Execution environment updated")
            self._pull_required = False
            self._assessment.pull_required = False
        except subprocess.CalledProcessError as exc:
            self._log_message(level=logging.ERROR, message="Execution environment pull failed")
            if exc.stderr is not None:
                self._log_message(level=logging.ERROR, message=exc.stderr.decode().strip())
            exit_msg = (
                "Check the execution environment image name, connectivity to and permissions"
                " for the registry, and try again"
            )
            self._log_message(level=logging.INFO, message=exit_msg, hint=True)



"""Version information for ansible-navigator collection doc cache.

Note:
    ``__version_collection_doc_cache__``:

    Indicates the version of the schema of the collection doc cache
    this is checked during initialization, if the version of the cache
    differs from below, the cache will be rebuilt.  This should be
    incremented when the schema changes and need not correspond to the
    application version, although keeping the major in sync is probably
    not a bad idea to minimize the amount of stale docs in the user's cache
"""

__version_collection_doc_cache__ = "1.1"



"""Object definitions related to utils."""

from __future__ import annotations

import logging
import textwrap

from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import NamedTuple


GOLDEN_RATIO = 1.61803398875


class Color:
    """Color constants."""

    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    GREY = "\033[90m"  # Bright black?
    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"
    END = "\033[0m"


class Decoration:
    """Decoration constants."""

    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    REVERSED = "\033[7m"
    END = "\033[0m"


class ExitPrefix(Enum):
    """An exit message prefix."""

    ERROR = "Error"
    HINT = "Hint"
    NOTE = "Note"
    WARNING = "Warning"

    @classmethod
    def _longest_name(cls) -> int:
        """Return the longest exit message prefix.

        Returns:
            The longest exit message prefix
        """
        return max(len(member.value) for member in cls)

    @classmethod
    def longest_formatted(cls) -> int:
        """Return the longest exit message prefix.

        Returns:
            The longest exit message prefix
        """
        return max(len(str(member)) for member in cls)

    def __str__(self) -> str:
        """Return the exit message prefix as a string.

        Returns:
            The exit message prefix as a string
        """
        name_len = self._longest_name()
        return f"{' ' * (name_len - len(self.name))}{self.name.capitalize()}: "


@dataclass
class ExitMessage:
    """An object to hold a message to present when exiting."""

    #: The message that will be presented
    message: str
    #: The prefix for the message, used for formatting
    prefix: ExitPrefix = ExitPrefix.ERROR

    @property
    def color(self) -> str:
        """Return a color for the prefix.

        Returns:
            The color for the prefix
        """
        color_mapping = {
            ExitPrefix.ERROR: Color.RED,
            ExitPrefix.HINT: Color.CYAN,
            ExitPrefix.NOTE: Color.GREEN,
            ExitPrefix.WARNING: Color.YELLOW,
        }
        return color_mapping[self.prefix]

    @property
    def level(self) -> int:
        """Return a log level.

        Returns:
            The log level
        """
        mapping = {
            ExitPrefix.ERROR: logging.ERROR,
            ExitPrefix.HINT: logging.INFO,
            ExitPrefix.NOTE: logging.INFO,
            ExitPrefix.WARNING: logging.WARNING,
        }
        return mapping[self.prefix]

    def to_lines(self, color: bool, width: int, with_prefix: bool) -> list[str]:
        """Output exit message to the console.

        Args:
            color: Whether to color the message
            width: Constrain message to width
            with_prefix: Whether to prefix the message

        Returns:
            The exit message as a string
        """
        prefix_length = ExitPrefix.longest_formatted()
        indent = " " * prefix_length

        message = textwrap.fill(
            self.message,
            width=width,
            break_on_hyphens=False,
            initial_indent=str(self.prefix) if with_prefix else indent,
            subsequent_indent=indent,
        )
        lines = message.splitlines()

        start_color = self.color if color else ""
        end_color = Color.END if color else ""

        printable = [f"{start_color}{line}{end_color}" for line in lines]
        return printable


@dataclass
class ExitMessages:
    """A mechanism to store multiple exit messages."""

    messages: list[ExitMessage] = field(default_factory=list)

    def to_strings(self, color: bool, width: int) -> list[str]:
        """Output exit messages to the console.

        Args:
            color: Whether to color the message
            width: Constrain messages to width

        Returns:
            The exit messages as a list of strings
        """
        printable = []
        new_section = True
        for idx, message in enumerate(self.messages):
            # Print the prefix if the next is different or a new section
            if new_section:
                # Use prefix, starting a new section
                with_prefix = True
            else:
                try:
                    # Prefix if previous is different
                    with_prefix = self.messages[idx - 1].prefix != message.prefix
                except IndexError:
                    # Last message
                    with_prefix = True
            printable.extend(message.to_lines(color=color, with_prefix=with_prefix, width=width))

            try:
                next_prefix = self.messages[idx + 1].prefix

                # Never break before a hint
                if next_prefix is ExitPrefix.HINT:
                    new_section = False
                    continue

                # Keep like items together
                if message.prefix is next_prefix:
                    new_section = False
                    continue

                # Start a new section
                printable.append("")
                new_section = True

            except IndexError:
                pass

        return printable


class LogMessage(NamedTuple):
    """An object to hold a message destined for the logger."""

    level: int
    message: str



"""Methods of transforming the settings."""

from __future__ import annotations

import json
import textwrap

from typing import Any

from ansible_navigator.utils.dict_merge import in_place_list_replace
from ansible_navigator.utils.functions import shlex_join
from ansible_navigator.utils.packaged_data import retrieve_content

from .definitions import ApplicationConfiguration
from .definitions import Constants
from .definitions import SettingsEntry
from .definitions import SettingsFileType
from .defs_presentable import PresentableSettingsEntries
from .defs_presentable import PresentableSettingsEntry
from .utils import create_settings_file_sample


def _convert_volume_mount(mount: str) -> dict[str, str]:
    """Convert a volume mount string to a dictionary.

    Args:
        mount: A colon-separated volume mount string.

    Returns:
        A dictionary with src, dest, and optionally options keys.
    """
    parts = mount.split(":")
    v_mount: dict[str, str] = {"src": parts[0], "dest": parts[1]}
    if len(parts) == 3:
        v_mount["options"] = parts[2]
    return v_mount


def to_effective(
    settings: ApplicationConfiguration,
) -> SettingsFileType:
    """Transform the current settings into a settings file.

    Args:
        settings: The current settings

    Returns:
        The settings represented as settings file
    """
    rebuilt: dict[Any, Any] = {}
    for entry in settings.entries:
        path = entry.settings_file_path(prefix=settings.application_name_dashed)
        if not isinstance(entry.value.current, Constants):
            current: bool | int | str | dict[Any, Any] | list[Any] = entry.value.current
            if path == "ansible-navigator.ansible.cmdline":
                current = shlex_join(entry.value.current)
            elif path == "ansible-navigator.execution-environment.volume-mounts":
                current = [_convert_volume_mount(mount) for mount in entry.value.current]

            partial = create_settings_file_sample(
                settings_path=path,
                placeholder=current,
            )
            in_place_list_replace(rebuilt, partial)
    return SettingsFileType(rebuilt)


def to_sources(settings: ApplicationConfiguration) -> dict[str, str]:
    """Transform the current settings into representation of sources.

    Args:
        settings: The current settings

    Returns:
        The settings sourced represented as a dictionary of path, source
    """
    sources = {}
    for entry in settings.entries:
        path = entry.settings_file_path(prefix=settings.application_name_dashed)
        sources[path] = str(entry.value.source)
    sources["settings_file_path"] = str(settings.internals.settings_file_path)
    sources["settings_file_source"] = str(settings.internals.settings_source)
    return sources


def to_presentable(settings: ApplicationConfiguration) -> PresentableSettingsEntries:
    """Transform the current settings into a structure that can be presented with the TUI.

    Args:
        settings: The current settings

    Returns:
        The settings represented as a list of dictionaries
    """
    settings_list = []

    all_subcommands = sorted(subcommand.name for subcommand in settings.subcommands)

    settings_file_entry = PresentableSettingsEntry.for_settings_file(
        all_subcommands=all_subcommands,
        application_name=settings.application_name_dashed,
        internals=settings.internals,
    )
    settings_list.append(settings_file_entry)

    settings_file_path = settings_file_entry.current_settings_file
    for entry in settings.entries:
        human_readable_entry = PresentableSettingsEntry.from_settings_entry(
            entry=entry,
            all_subcommands=all_subcommands,
            application_name_dashed=settings.application_name_dashed,
            settings_file_path=settings_file_path,
        )
        settings_list.append(human_readable_entry)

    settings_list.sort()
    return PresentableSettingsEntries(tuple(settings_list))


def _apply_entry_to_schema(
    entry: SettingsEntry,
    subschema: dict[Any, Any],
    dot_parts: list[str],
) -> None:
    """Apply a single entry's metadata to the schema.

    Args:
        entry: The settings entry.
        subschema: The subschema dict to populate.
        dot_parts: The dot-separated path parts.
    """
    subschema[dot_parts[-1]]["description"] = entry.short_description
    if entry.choices:
        choices = list(entry.choices)
        if subschema[dot_parts[-1]].get("type") == "array":
            subschema[dot_parts[-1]]["items"]["enum"] = choices
        else:
            subschema[dot_parts[-1]]["enum"] = choices
    if entry.value.schema_default is not Constants.NOT_SET:
        if entry.value.schema_default is not Constants.NONE:
            subschema[dot_parts[-1]]["default"] = entry.value.schema_default
    elif entry.value.default is not Constants.NOT_SET:
        subschema[dot_parts[-1]]["default"] = entry.value.default


def _resolve_schema_version(settings: ApplicationConfiguration) -> str:
    """Resolve the application version for schema use.

    Args:
        settings: The application settings.

    Returns:
        The major version string.
    """
    if isinstance(settings.application_version, Constants):
        version = settings.application_version.value
    else:
        version = settings.application_version
    return ".".join(version.split(".")[:1])


def to_schema(settings: ApplicationConfiguration) -> dict[str, Any]:
    """Build a json schema from the settings using the stub schema.

    Args:
        settings: The application settings

    Returns:
        The json schema
    """
    file_contents = retrieve_content("settings-schema.partial.json")
    partial_schema = json.loads(file_contents)

    for entry in settings.entries:
        subschema: dict[Any, Any] = partial_schema["properties"]
        dot_parts = entry.settings_file_path(prefix=settings.application_name_dashed).split(".")
        for part in dot_parts[:-1]:
            if isinstance(subschema, dict):
                subschema = subschema.get(part, {}).get("properties")
        _apply_entry_to_schema(entry, subschema, dot_parts)

    version = _resolve_schema_version(settings)
    partial_schema["version"] = version
    partial_schema["title"] = partial_schema["title"].format(version=version)

    return partial_schema


def to_sample(settings: ApplicationConfiguration) -> tuple[str, str]:
    """Load and clean the settings sample.

    Args:
        settings: The application settings

    Returns:
        One selectively commented sample, one uncommented
    """
    # pylint: disable=too-many-locals

    file_contents = retrieve_content(filename="settings-sample.template.yml").splitlines()

    # Remove anything before the `---`
    yaml_doc_start = file_contents.index("---")

    template_lines = file_contents[yaml_doc_start:]

    # Find all anchors
    indices: list[tuple[str, int, SettingsEntry]] = []
    for entry in settings.entries:
        dot_path = entry.settings_file_path(prefix="")
        indent = "  " * len(dot_path.split("."))  # indent 2 spaces for each part
        comment_index = template_lines.index(f"{indent}# {{{{ {dot_path} }}}}")
        indices.append((indent, comment_index, entry))

    # Replace anchors with the short description
    sorted_entries = sorted(indices, key=lambda t: -t[1])
    for indent, comment_index, entry in sorted_entries:
        description = [f"{indent}# {line}" for line in textwrap.wrap(entry.short_description)]
        template_lines = (
            template_lines[0:comment_index] + description + template_lines[comment_index + 1 :]
        )

    populated_lines = template_lines

    no_comment = ["---", "ansible-navigator:", "logging:", "level: debug", "append: False"]
    commented_lines = []
    for line in populated_lines:
        if any(entry in line for entry in no_comment):
            commented_lines.append(line)
        else:
            commented_lines.append(f"# {line}")

    commented = "\n".join(commented_lines) + "\n"
    uncommented = "\n".join(populated_lines) + "\n"

    return commented, uncommented



"""Introspect an execution environment image."""

from __future__ import annotations

import argparse
import json
import os
import re
import subprocess
import sys
import threading

from queue import Empty
from queue import Queue
from types import SimpleNamespace
from typing import TYPE_CHECKING
from typing import Any
from typing import TypeAlias


if TYPE_CHECKING:
    from collections.abc import Callable

_WORKER_TIMEOUT = 30

# https://github.com/python/typing/issues/182#issuecomment-1320974824
JSONTypes: TypeAlias = dict[str, "JSONTypes"] | list["JSONTypes"] | str | int | float | bool | None


class Command(SimpleNamespace):
    """Abstraction for a details about a shell command."""

    id_: str
    command: str
    parse: Callable[..., Any]
    stdout: str = ""
    stderr: str = ""
    details: list[str] | dict[Any, Any] | list[dict[Any, Any]] | str = ""
    errors: list[str] = []


def run_command(command: Command) -> None:
    """Run a command using subprocess.

    Args:
        command: Details of the command to run
    """
    try:
        proc_out = subprocess.run(
            command.command,
            capture_output=True,
            check=True,
            text=True,
            shell=True,
        )
        command.stdout = proc_out.stdout
    except subprocess.CalledProcessError as exc:
        command.stderr = str(exc.stderr)
        command.errors = [str(exc.stderr)]


def worker(pending_queue: Queue[Any], completed_queue: Queue[Any]) -> None:
    """Run a command from pending, parse, and place in completed.

    Args:
        pending_queue: A queue with plugins to process
        completed_queue: The queue in which extracted documentation will
            be placed
    """
    while True:
        command = pending_queue.get()
        if command is None:
            break
        run_command(command)
        try:
            command.parse(command)
        except Exception as exc:  # noqa: BLE001
            command.errors = command.errors + [str(exc)]
        completed_queue.put(command)


class CommandRunner:
    """A command runner.

    Run commands using single or multiple processes.
    """

    def __init__(self) -> None:
        """Initialize the command runner."""
        self._completed_queue: Queue[Any] | None = None
        self._pending_queue: Queue[Any] | None = None

    def run_multi_thread(self, command_classes: list[CmdParser]) -> list[CmdParser]:
        """Run commands with multiple threads.

        Workers are started to read from pending queue.
        Exit when the number of results is equal to the number
        of commands needing to be run.

        Args:
            command_classes: All command classes to be run

        Returns:
            The results from running all commands

        Raises:
            RuntimeError: If all workers terminate before all commands complete
        """
        if self._completed_queue is None:
            self._completed_queue = Queue()
        if self._pending_queue is None:
            self._pending_queue = Queue()
        all_commands = tuple(
            command for command_class in command_classes for command in command_class.commands
        )
        threads = self.start_workers(all_commands)
        results: list[CmdParser] = []
        while len(results) != len(all_commands):
            try:
                results.append(self._completed_queue.get(timeout=_WORKER_TIMEOUT))
            except Empty:
                alive_threads = [t for t in threads if t.is_alive()]
                if not alive_threads:
                    failed = len(all_commands) - len(results)
                    msg = (
                        f"All worker threads have terminated but {failed} of "
                        f"{len(all_commands)} commands did not complete. "
                        f"Worker threads may have crashed."
                    )
                    raise RuntimeError(msg)  # noqa: B904
        for t in threads:
            t.join()
        return results

    def start_workers(self, jobs: tuple[Command, ...]) -> list[threading.Thread]:
        """Start workers and submit jobs to pending queue.

        Args:
            jobs: The jobs to be run

        Returns:
            The list of started worker threads

        Raises:
            RuntimeError: if the pending queue is not initialized
        """
        worker_count = len(jobs)
        threads: list[threading.Thread] = []
        for _proc in range(worker_count):
            proc = threading.Thread(
                target=worker,
                args=(self._pending_queue, self._completed_queue),
            )
            threads.append(proc)
            proc.start()
        if not self._pending_queue:
            raise RuntimeError
        for job in jobs:
            self._pending_queue.put(job)
        for _proc in range(worker_count):
            self._pending_queue.put(None)
        return threads


class CmdParser:
    """A base class for command parsers with common parsing functions."""

    @property
    def commands(self) -> list[Command]:
        """List of commands to be executed."""
        return []

    @staticmethod
    def _strip(value: str) -> str:
        """Remove quotes, leading and trailing whitespace.

        Args:
            value: The string to act on

        Returns:
            The string after removing quotes, leading and trailing
            whitespace
        """
        return value.strip('"').strip("'").strip()

    @staticmethod
    def re_partition(content: Any, separator: str) -> Any:
        """Partition a string using a regular expression.

        Args:
            content: The content to partition
            separator: The separator to use for the partitioning

        Returns:
            The first partition, separator, and final partition
        """
        separator_match = re.search(separator, content)
        if not separator_match or content.startswith(" "):
            return "", "", content
        delim = separator_match.group(0)
        key, content = re.split(delim, content, maxsplit=1)
        return key, delim, content

    def splitter(
        self,
        lines: list[str],
        line_split: str,
        section_delim: str | None = None,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """Split lines given a delimiter.

        Args:
            lines: The lines to split
            line_split: The delimiter use for splitting each line
            section_delim: The separator between different packages

        Returns:
            All lines split on the delimiter
        """
        results: list[dict[str, Any]] = []
        result: dict[str, Any] = {}
        current_key: str = ""
        while lines:
            line = lines.pop(0)
            key, delim, content = self.re_partition(line, line_split)
            content = self._strip(content)
            if section_delim and line == section_delim:
                results.append(result)
                result = {}
                continue
            if not delim and current_key:
                result[current_key] += f" {content}"
                continue
            current_key = key.lower().replace("_", "-").strip()
            # system_packages description field needs special handling
            if current_key == "description":
                description = []
                while lines:
                    line = lines.pop(0)
                    if line == section_delim:
                        break
                    description.append(line)
                if description:
                    result[current_key] = " ".join(description)
                else:
                    result[current_key] = "No description available"
                results.append(result)
                return result
            result[current_key] = content

        if result:
            results.append(result)
        return results


class AnsibleCollections(CmdParser):
    """Available ansible collections collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the ansible-galaxy command to list ansible collections.

        Returns:
            The defined command
        """
        command = "ansible-galaxy collection list"
        return [
            Command(
                id_="ansible_collections",
                command=command,
                parse=self.parse,
            ),
        ]

    @staticmethod
    def parse(command: Command) -> None:
        """Parse the output of the ansible-galaxy command.

        Args:
            command: The result of running the command
        """
        if "invalid choice: 'list'" in command.stderr:
            command.details = "This command is not supported with ansible 2.9."
            return
        collections = {}
        for line in command.stdout.splitlines():
            parts = line.split()
            if len(parts) == 2 and parts[1][0].isdigit():
                collections[parts[0].strip()] = parts[1].strip()
        command.details = collections


class AnsibleVersion(CmdParser):
    """Ansible version collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the ansible command to get the version.

        Returns:
            The defined command
        """
        return [Command(id_="ansible_version", command="ansible --version", parse=self.parse)]

    @staticmethod
    def parse(command: Command) -> None:
        """Parse the output of the ansible command.

        Args:
            command: The result of running the command
        """
        version = command.stdout.splitlines()[0]
        command.details = version


class OsRelease(CmdParser):
    """OS release information collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the command to collect os release information.

        Returns:
            The defined command
        """
        return [Command(id_="os_release", command="cat /etc/os-release", parse=self.parse)]

    def parse(self, command: Command) -> None:
        """Parse the output of the cat command.

        Args:
            command: The result of running the command
        """
        parsed = self.splitter(command.stdout.splitlines(), "=")
        command.details = parsed


class PythonPackages(CmdParser):
    """Python package collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the pip command to list installed pip packages.

        Returns:
            The defined command
        """
        pre = Command(
            id_="pip_freeze",
            command=f"{sys.executable} -m pip freeze",
            parse=self.parse_freeze,
        )
        run_command(pre)
        pre.parse(pre)
        pkgs = " ".join(pre.details[0]) if pre.details else ""
        return [
            Command(
                id_="python_packages",
                command=f"{sys.executable} -m pip show {pkgs}",
                parse=self.parse,
            ),
        ]

    def parse(self, command: Command) -> None:
        """Parse the output of the pip command.

        Args:
            command: The result of running the command
        """
        parsed = self.splitter(command.stdout.splitlines(), line_split=":", section_delim="---")
        if isinstance(parsed, dict):
            parsed = [parsed]
        for pkg in parsed:
            for entry in ["required-by", "requires"]:
                if pkg[entry]:
                    pkg[entry] = [p.strip() for p in pkg[entry].split(",")]
                else:
                    pkg[entry] = []
        command.details = parsed

    def parse_freeze(self, command: Command) -> None:
        """Parse the output of the pip freeze command, skipping editables.

        Args:
            command: The result of running the command
        """
        lines = [line for line in command.stdout.splitlines() if not line.startswith("-e")]
        parsed = self.splitter(lines, "(==|@)")
        command.details = parsed


class RedhatRelease(CmdParser):
    """Red Hat release collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the command to get the redhat release information.

        Returns:
            The defined command
        """
        return [Command(id_="redhat_release", command="cat /etc/redhat-release", parse=self.parse)]

    @staticmethod
    def parse(command: Command) -> None:
        """Parse the output of the cat redhat release command.

        Args:
            command: The result of running the command
        """
        parsed = command.stdout
        command.details = parsed


class SystemPackages(CmdParser):
    """System packages collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the command to list system packages.

        Returns:
            The defined command
        """
        return [Command(id_="system_packages", command="rpm -qai", parse=self.parse)]

    def parse(self, command: Command) -> None:
        """Parse the output of the rpm command.

        Args:
            command: The result of running the command
        """
        packages = []
        package: list[str] = []
        for line in command.stdout.splitlines():
            if re.match(r"^Name\s{2,}:", line) and package:
                packages.append(package)
                package = [line]
            else:
                package.append(line)
        if package:
            packages.append(package)

        parsed: list[Any] = []
        for package in packages:
            result = self.splitter(package, line_split=":")
            parsed.append(result)

        command.details = parsed


ALL_COLLECTORS: dict[str, type[CmdParser]] = {
    "ansible_collections": AnsibleCollections,
    "ansible_version": AnsibleVersion,
    "os_release": OsRelease,
    "redhat_release": RedhatRelease,
    "python_packages": PythonPackages,
    "system_packages": SystemPackages,
}


def main(
    serialize: bool = True,
    sections: list[str] | None = None,
) -> dict[str, JSONTypes] | None:
    """Enter the image introspection process.

    Args:
        serialize: Whether to serialize the results
        sections: Optional list of section IDs to collect. When ``None`` or
            when ``"everything"`` is present, all collectors run.

    Returns:
        The collected data or none if serialize is False
    """
    response: dict[str, Any] = {"errors": []}
    response["python_version"] = {"details": {"version": " ".join(sys.version.splitlines())}}
    response["environment_variables"] = {"details": dict(os.environ)}
    try:
        command_runner = CommandRunner()

        if sections and "everything" not in sections:
            commands: list[CmdParser] = [
                ALL_COLLECTORS[s]() for s in sections if s in ALL_COLLECTORS
            ]
        else:
            commands = [cls() for cls in ALL_COLLECTORS.values()]

        results = command_runner.run_multi_thread(commands)
        for result in results:
            result_as_dict = vars(result)
            result_as_dict.pop("parse")
            for key in list(result_as_dict):
                if key not in ["details", "errors"]:
                    result_as_dict[f"__{key}"] = result_as_dict[key]
                    result_as_dict.pop(key)
            response[result_as_dict["__id_"]] = result_as_dict
    except Exception as exc:  # noqa: BLE001
        response["errors"].append(str(exc))
    if serialize:
        print(json.dumps(response))
        return None
    return response


ALWAYS_COLLECTED = ("python_version", "environment_variables")


def _parse_args() -> list[str] | None:
    """Parse command-line arguments for section filtering.

    Returns:
        The list of requested sections, or None for all sections.
    """
    parser = argparse.ArgumentParser(description="Introspect an execution environment image.")
    parser.add_argument(
        "--sections",
        nargs="*",
        default=None,
        choices=[*ALL_COLLECTORS, *ALWAYS_COLLECTED, "everything"],
        help="Sections to collect. Omit for all sections.",
    )
    args = parser.parse_args()
    return args.sections


if __name__ == "__main__":  # pragma: no cover
    requested_sections = _parse_args()
    main(sections=requested_sections)



"""Some UI specific utils."""

from __future__ import annotations

import functools
import re

from dataclasses import is_dataclass
from math import floor
from typing import TYPE_CHECKING
from typing import Any


if TYPE_CHECKING:
    from ansible_navigator.content_defs import ContentBase


def convert_percentage(
    content: dict[str, Any] | ContentBase[Any],
    columns: list[str],
    progress_bar_width: int,
) -> None:
    """Convert an attribute or value to a progress bar formatted for the TUI in place.

    Args:
        content: The content for which the progress bar will be
            generated
        columns: The menu columns, only make progress bars for columns
        progress_bar_width: The target width of the progress bar
    """
    for column in columns:
        value = content.get(column)
        if value and is_percent(str(value)):
            new_value = _string_to_progress(value, progress_bar_width)
            if isinstance(content, dict):
                content["_" + column] = value
                content[column] = new_value
            elif is_dataclass(content):
                setattr(content, f"_{column}", value)
                setattr(content, column, new_value)


def _string_to_progress(value: str, progress_bar_width: int) -> str:
    """Convert a string to a progress bar or text string indicating complete.

    Args:
        value: The percent string
        progress_bar_width: The target width of the progress bar

    Returns:
        The resulting progress bar or text string
    """
    if value == "100%":
        return "Complete".center(progress_bar_width, " ")
    chars_in_progress_bar = floor(progress_bar_width / 100 * int(value[0:-1]))
    return ("\u2587" * chars_in_progress_bar).ljust(progress_bar_width)


@functools.cache
def is_percent(string: str) -> bool:
    """Determine if a string is a percentage.

    Args:
        string: The string to check

    Returns:
        Boolean of comparison
    """
    if string.endswith("%") and re.match(r"^\d{1,3}%$", string):  # noqa:SIM103
        return True
    return False


def distribute(available: int, weights: list[int]) -> list[int]:
    """Distribute some available fairly across a list of numbers.

    Args:
        available (int): the total
        weights (List[int]): numbers

    Returns:
        List of distributed amounts available
    """
    total = sum(weights)
    if available < total:
        while available != total:
            max_value = max(weights)
            max_values = [i for i, j in enumerate(weights) if j == max_value]
            for idx in max_values:
                weights[idx] -= 1
                if sum(weights) == available:
                    return weights

    distributed_amounts = []
    total_weights = sum(weights)
    for weight in weights:
        percent_of_total = float(weight) / total_weights
        distributed_amount = round(percent_of_total * available)
        distributed_amounts.append(distributed_amount)
        total_weights -= weight
        available -= distributed_amount
    return distributed_amounts



"""``:back`` command implementation.

Additionally triggered by the escape key.
"""

import logging

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action:
    """``:back`` command implementation."""

    KEGEX = r"^\^\[|\x1b|back$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:back`` action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the ``:back`` request.

        Args:
            interaction: The interaction from the user
            app: The app instance
        """
        self._logger.debug("back requested")
        interaction.ui.scroll(0)
        interaction.ui.clear()
        this = app.steps.back_one()  # pop this
        step = app.steps.back_one()  # pop current

        if (
            app.steps
            and isinstance(step, Step)
            and isinstance(app.steps.current, Step)
            and step.type == "menu"
            and app.steps.current.type == "menu"
        ):
            interaction.ui.menu_filter(None)
            self._logger.debug(
                "Stepping back in %s from %s to %s",
                app.name,
                step.name,
                app.steps.current.name,
            )
        else:
            self._logger.debug("Return to %s, at last step", app.name)

        app.steps.append(this)  # put this back on



"""A carrier for app internals.

This will be shared with other actions and is immutable.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import NamedTuple


if TYPE_CHECKING:
    from collections.abc import Callable

    from .configuration_subsystem.definitions import ApplicationConfiguration
    from .steps import Steps


class AppPublic(NamedTuple):
    """A carrier class for app internals.

    This will be shared with other actions and is immutable.
    """

    # Quoted due to https://github.com/sphinx-doc/sphinx/issues/10400
    args: ApplicationConfiguration
    name: str
    rerun: Callable[[], None]
    stdout: list[str]
    steps: Steps
    update: Callable[..., None]
    write_artifact: Callable[..., None]



"""Information handler, instant enter."""

from __future__ import annotations

from curses import ascii as curses_ascii
from typing import TYPE_CHECKING
from typing import Any

from .curses_window import CursesWindow
from .curses_window import Window


if TYPE_CHECKING:
    from ansible_navigator.ui_framework.ui_config import UIConfig

    from .field_information import FieldInformation


class FormHandlerInformation(CursesWindow):
    """Handle form button."""

    def __init__(self, screen: Window, ui_config: UIConfig) -> None:
        """Initialize the handler for a informational notification.

        Args:
            screen: A curses window
            ui_config: The current user interface configuration
        """
        super().__init__(ui_config=ui_config)
        self._screen = screen

    @staticmethod
    def handle(idx: int, form_fields: list[Any]) -> tuple[FieldInformation, int]:
        """Handle the information field, immediate return.

        Args:
            idx: Index to retrieve specific field
            form_fields: List of fields

        Returns:
            Indexed form fields
        """
        return form_fields[idx], curses_ascii.NL



"""A couple singleton sentinels for convenience to avoid the use of None."""

from __future__ import annotations

from typing import Any


class Singleton(type):
    """One of a kind."""

    _instances: dict[Any, Singleton] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        """Determine if the instance is one of a kind.

        Args:
            *args: Arbitrary arguments
            **kwargs: Dict of Keyword Args

        Returns:
            The type
        """
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Unknown(metaclass=Singleton):
    """Something that should eventually be known."""

    def __repr__(self) -> str:
        """Return the type of the object.

        Returns:
            The type of the object.
        """
        return type(self).__name__


unknown = Unknown()


class Nonexistent(metaclass=Singleton):
    """Something that does not exist."""

    def __repr__(self) -> str:
        """Return the type of the object.

        Returns:
            The type of the object
        """
        return type(self).__name__


nonexistent = Nonexistent()



"""The configuration subsystem for Ansible Navigator."""

from .configurator import Configurator
from .definitions import Constants
from .definitions import SettingsEntry
from .definitions import SettingsFileType
from .definitions import SettingsSchemaType
from .defs_presentable import PresentableSettingsEntries
from .defs_presentable import PresentableSettingsEntry
from .navigator_configuration import NavigatorConfiguration
from .transform import to_effective
from .transform import to_presentable
from .transform import to_sample
from .transform import to_schema
from .transform import to_sources


__all__ = (
    "Configurator",
    "Constants",
    "NavigatorConfiguration",
    "PresentableSettingsEntries",
    "PresentableSettingsEntry",
    "SettingsEntry",
    "SettingsFileType",
    "SettingsSchemaType",
    "to_effective",
    "to_presentable",
    "to_sample",
    "to_schema",
    "to_sources",
)



"""Object to hold basic UI settings."""

from dataclasses import dataclass

from ansible_navigator.utils.compatibility import Traversable


@dataclass
class UIConfig:
    """Object to hold basic UI settings.

    Used to determine properties about rendering things. An instance of this
    class gets threaded throughout most of the UI system, so it can be used for
    fairly global things, such as "should we render color, ever?"
    """

    #: Indicates coloring is enabled or disabled
    color: bool
    #: Indicates if the curses colors have been initialized
    colors_initialized: bool
    #: The path to the grammar directory
    grammar_dir: Traversable
    #: Indicates if terminal support for OSC4 is enabled
    osc4: bool
    #: The path to the 16 terminal color map
    terminal_colors_path: Traversable
    #: The path to the theme file
    theme_path: Traversable



"""Type for curses window."""

from __future__ import annotations

import curses
import json
import logging

from typing import TYPE_CHECKING

from .colorize import hex_to_rgb_curses


if TYPE_CHECKING:
    from _curses import window

    from .curses_defs import CursesLine
    from .curses_defs import CursesLinePart
    from .ui_config import UIConfig

    Window = window
else:
    from typing import Any

    Window = Any


COLOR_MAP = {
    "terminal.ansiBlack": 0,
    "terminal.ansiRed": 1,
    "terminal.ansiGreen": 2,
    "terminal.ansiYellow": 3,
    "terminal.ansiBlue": 4,
    "terminal.ansiMagenta": 5,
    "terminal.ansiCyan": 6,
    "terminal.ansiWhite": 7,
    "terminal.ansiBrightBlack": 8,
    "terminal.ansiBrightRed": 9,
    "terminal.ansiBrightGreen": 10,
    "terminal.ansiBrightYellow": 11,
    "terminal.ansiBrightBlue": 12,
    "terminal.ansiBrightMagenta": 13,
    "terminal.ansiBrightCyan": 14,
    "terminal.ansiBrightWhite": 15,
}


class CursesWindow:
    # pylint: disable=too-many-instance-attributes
    """Abstraction for a curses window."""

    def __init__(self, ui_config: UIConfig) -> None:
        """Initialize a curses window.

        Args:
            ui_config: The current user interface configuration
        """
        self._logger = logging.getLogger(__name__)

        self._screen: Window
        self.win: Window
        self._screen_min_height = 3
        self._prefix_color = 8
        self._theme_dir: str
        self._term_osc4_support: bool
        self._ui_config = ui_config
        self._logger.debug("self._ui_config: %s", self._ui_config)
        self._set_colors()

    @property
    def _screen_width(self) -> int:
        """Return the screen width.

        Returns:
            the current screen width
        """
        return self._screen.getmaxyx()[1]

    @property
    def _screen_height(self) -> int:
        """Return the screen height, or notify if too small.

        Returns:
            the current screen height
        """
        while True:
            if self._screen.getmaxyx()[0] >= self._screen_min_height:
                return self._screen.getmaxyx()[0]
            curses.flash()
            curses.beep()
            self._screen.refresh()

    def _color_pair_or_none(self, color: int) -> int | None:
        """Return 0 if colors are disabled, otherwise returns the curses color pair.

        Args:
            color: Int for specific curses color

        Returns:
            Curses color pair
        """
        if not self._ui_config.color or curses.COLORS == 0:
            return None
        color_arg = color % curses.COLORS  # self._number_colors
        return curses.color_pair(color_arg)

    def _curs_set(self, value: int) -> None:
        """Log an error in the case of a TERM with limited capabilities.

        Args:
            value: TERM value
        """
        try:
            curses.curs_set(value)
        except curses.error:
            self._logger.exception("Errors setting up terminal, check TERM value")

    def _render_line_part(
        self,
        window: Window,
        lineno: int,
        line_part: CursesLinePart,
        prefix_len: int,
    ) -> None:
        """Render a single line part to a curses window.

        Args:
            window: A curses window
            lineno: the line number
            line_part: The individual line part to render
            prefix_len: The length of the line prefix
        """
        column = line_part.column + prefix_len
        if column > self._screen_width:
            return

        text = line_part.string[0 : self._screen_width - column + 1]
        try:
            color = self._color_pair_or_none(line_part.color)
            if color is None:
                window.addstr(lineno, column, text)
            else:
                window.addstr(lineno, column, text, color | line_part.decoration)
        except curses.error:
            # curses error at last column & row but I don't care
            # because it still draws it
            # https://stackoverflow.com/questions/10877469/
            # ncurses-setting-last-character-on-screen-without-scrolling-enabled
            if lineno == window.getyx()[0] and column + len(text) == window.getyx()[1] + 1:
                pass
            else:
                self._logger.debug("curses error")
                self._logger.debug(
                    "screen_height: %s, lineno: %s",
                    self._screen_height,
                    lineno,
                )
                self._logger.debug(
                    "screen_w: %s, column: %s text: %s, lentext: %s, end_col: %s",
                    self._screen_width,
                    column,
                    text,
                    len(text),
                    column + len(text),
                )

    def _add_line(
        self,
        window: Window,
        lineno: int,
        line: CursesLine,
        prefix: str | None = None,
    ) -> None:
        """Add a line to a window.

        Args:
            window (Window): A curses window
            lineno (int): the line number
            line (CursesLine): The line to add
            prefix (str or None): The prefix for the line
        """
        if prefix:
            color = self._color_pair_or_none(self._prefix_color)
            if color is None:
                window.addstr(lineno, 0, prefix)
            else:
                window.addstr(lineno, 0, prefix, color)
        if line:
            window.move(lineno, 0)
            for line_part in line:
                self._render_line_part(window, lineno, line_part, len(prefix or ""))

    def _set_colors(self) -> None:
        """Set the colors for curses."""
        # curses colors may have already been initialized
        # with another instance of curses window
        if self._ui_config.colors_initialized is True:
            return

        self._curs_set(0)
        # in the case of a TERM with limited capabilities
        # disable color and get out fast
        try:
            curses.use_default_colors()
        except curses.error:
            self._logger.exception("Errors setting up terminal, no color support")
            self._term_osc4_support = False
            self._ui_config.colors_initialized = True
            return

        self._logger.debug("curses.COLORS: %s", curses.COLORS)
        self._logger.debug("curses.can_change_color: %s", curses.can_change_color())

        self._term_osc4_support = curses.can_change_color()
        if self._ui_config.osc4 is False:
            self._term_osc4_support = False
        self._logger.debug("term_osc4_support: %s", self._term_osc4_support)

        if self._term_osc4_support:
            with self._ui_config.terminal_colors_path.open("r", encoding="utf-8") as fh:
                colors = json.load(fh)

            for color_name, color_hex in colors.items():
                idx = COLOR_MAP[color_name]
                color = hex_to_rgb_curses(color_hex)
                curses.init_color(idx, *color)
            self._logger.debug("Custom colors set")
        else:
            self._logger.debug("Using terminal defaults")

        for i in range(curses.COLORS):
            curses.init_pair(i, i, -1)
        self._ui_config.colors_initialized = True



"""An information field formatted as curses lines."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any

from .form_handler_information import FormHandlerInformation
from .sentinels import Unknown
from .sentinels import unknown
from .validators import FieldValidators


if TYPE_CHECKING:
    from collections.abc import Callable

    from .curses_defs import CursesLines
    from .curses_window import Window


@dataclass
class FieldCursesInformation:
    """An information field made of curses lines."""

    name: str
    information: CursesLines
    current_error: str = ""
    window_handler = FormHandlerInformation
    valid: bool | Unknown = unknown
    validator: Callable[..., Any] = FieldValidators.null
    win: Window | None = None

    @property
    def full_prompt(self) -> str:
        """Return the max width string.

        Returns:
            The max width string
        """
        return max(
            (max(line, key=lambda lp: len(lp.string)).string for line in self.information),
            key=len,
        )

    def validate(self, response: str) -> None:
        """No validation required for information field.

        Args:
            response: The response to validate
        """
        self.valid = True

    def conditional_validation(self, response: str) -> None:
        """No conditional validation.

        Args:
            response: The response to validate
        """
        self.validate(response)
